# -------------------------------------------------------------------------
# Copyright (c) PTC Inc. and/or all its affiliates. All rights reserved.
# See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------


r"""`error` Exception classes raised by Kepconfig.
Includes KepURLError and KepHTTPError
"""

__all__ = ['KepError', 'KepURLError', 'KepHTTPError']


class KepError(Exception):
    '''General Exception class for Kepconfig.
    '''
    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
            return 'KepError Error: %s' % (self.msg)

class KepURLError(KepError):
    '''Exception class raised by Kepconfig that inherits responses from the urllib URLError exceptions.
    '''
    def __init__(self, reason, url):
        KepError.__init__(self)
        self.reason = reason
        self.url = url

    def __str__(self):
        return '<urlopen error %s>' % self.reason

class KepHTTPError(KepError):
    '''Exception class raised by Kepconfig that inherits responses from the urllib HTTPError 
    exceptions. This exception class is also a valid HTTP response instance.  It behaves 
    this way because HTTP protocol errors are valid responses, with a status 
    code, headers, and a body.  In some contexts,an application may want to 
    handle an exception like a regular response.
    '''
    def __init__(self, url, code, msg, hdrs, payload):
        KepError.__init__(self, msg)
        self.url = url
        self.code = code
        self.hdrs = hdrs
        self.payload = payload
    
    @property
    def reason(self):
        return self.msg

    def __str__(self):
        return 'HTTP Error %s: %s' % (self.code, self.msg)


# -------------------------------------------------------------------------
# Copyright (c) 2020, PTC Inc. and/or all its affiliates. All rights reserved.
# See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------


__path__ = __import__("pkgutil").extend_path(__path__, __name__)
from . import exchange, range, name

CONSUMER_EXCHANGE = 'CONSUMER'
PRODUCER_EXCHANGE = 'PRODUCER'


# -------------------------------------------------------------------------
# Copyright (c) PTC Inc. and/or all its affiliates. All rights reserved.
# See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------


r"""`device` exposes an API to allow modifications (add, delete, modify) to 
device objects within the Kepware Configuration API
"""

from ..connection import KepServiceResponse
from ..error import KepHTTPError, KepError
from typing import Union
import kepconfig
from . import channel, tag
import inspect

DEVICE_ROOT = '/devices'
ATG_URL = '/services/TagGeneration'

def _create_url(device = None):
    '''Creates url object for the "device" branch of Kepware's project tree. Used 
    to build a part of Kepware Configuration API URL structure
    
    Returns the device specific url when a value is passed as the device name.
    '''
    if device == None:
        return DEVICE_ROOT
    else:
        return '{}/{}'.format(DEVICE_ROOT,device)

def add_device(server, dev_channel, DATA) -> Union[bool, list]:
    '''Add a "device" object to a channel in Kepware. Can be used to pass children of a device object 
    such as tags and tag groups. This allows you to create a device and tags 
    all in one function, if desired.

    Additionally it can be used to pass a list of channelsdevices and it's children to be added all at once.

    INPUTS:
    "server" - instance of the "server" class

    "dev_channel" - channel the device object exists

    "DATA" - properly JSON object (dict) of the device and it's children 
    expected by Kepware Configuration API
    
    RETURNS:
    True - If a "HTTP 201 - Created" is received from Kepware
    
    List  - If a "HTTP 207 - Multi-Status" is received from Kepware with a list of dict error responses for all 
    devices added that failed.

    EXCEPTIONS:
    KepHTTPError - If urllib provides an HTTPError
    KepURLError - If urllib provides an URLError
    '''

    r = server._config_add(server.url + channel._create_url(dev_channel) + _create_url(), DATA)
    if r.code == 201: return True
    elif r.code == 207:
        errors = [] 
        for item in r.payload:
            if item['code'] != 201:
                errors.append(item)
        return errors
    else: 
        raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)

def del_device(server, device_path) -> bool:
    '''Delete a "device" object in Kepware. This will delete all children as well.

    INPUTS:
    "server" - instance of the "server" class

    "device_path" - path identifying device to delete. Standard Kepware address decimal notation string including the 
    device such as "channel1.device1"

    RETURNS:
    True - If a "HTTP 200 - OK" is received from Kepware

    EXCEPTIONS:
    KepHTTPError - If urllib provides an HTTPError
    KepURLError - If urllib provides an URLError
    '''

    path_obj = kepconfig.path_split(device_path)
    try:
        r = server._config_del(server.url + channel._create_url(path_obj['channel']) + _create_url(path_obj['device']))
        if r.code == 200: return True 
        else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)
    except KeyError as err:
        err_msg = 'Error: No {} identified in {} | Function: {}'.format(err,'device_path', inspect.currentframe().f_code.co_name)
        raise KepError(err_msg)

def modify_device(server, device_path, DATA, force = False) -> bool:
    '''Modify a device object and it's properties in Kepware.

    INPUTS:
    "server" - instance of the "server" class

    "device_path" -  path identifying device to modify. Standard Kepware address decimal notation string including the 
    device such as "channel1.device1"

    "DATA" - properly JSON object (dict) of the device properties to be modified.

    "force" (optional) - if True, will force the configuration update to the Kepware server
    
    RETURNS:
    True - If a "HTTP 200 - OK" is received from Kepware

    EXCEPTIONS:
    KepHTTPError - If urllib provides an HTTPError
    KepURLError - If urllib provides an URLError
    '''

    device_data = server._force_update_check(force, DATA)

    path_obj = kepconfig.path_split(device_path)
    try:
        r = server._config_update(server.url + channel._create_url(path_obj['channel']) + _create_url(path_obj['device']), device_data)
        if r.code == 200: return True 
        else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)
    except KeyError as err:
            err_msg = 'Error: No {} identified in {} | Function: {}'.format(err,'device_path', inspect.currentframe().f_code.co_name)
            raise KepError(err_msg)
    # except Exception as e:
    #     return 'Error: Error with {}: {}'.format(inspect.currentframe().f_code.co_name, str(e))

def get_device(server, device_path) -> dict:
    '''Returns the properties of the device object. Returned object is JSON.

    INPUTS:
    "server" - instance of the "server" class
    
    "device_path" -  path identifying device. Standard Kepware address decimal notation string including the 
    device such as "channel1.device1"

    RETURNS:
    dict - data for the device requested

    EXCEPTIONS:
    KepHTTPError - If urllib provides an HTTPError
    KepURLError - If urllib provides an URLError
    '''

    path_obj = kepconfig.path_split(device_path)
    try:
        r = server._config_get(server.url + channel._create_url(path_obj['channel']) + _create_url(path_obj['device']))
        return r.payload
    except KeyError as err:
        err_msg = 'Error: No {} identified in {} | Function: {}'.format(err,'device_path', inspect.currentframe().f_code.co_name)
        raise KepError(err_msg)
    # except Exception as err:
    #     print('Error: Error with {}: {}'.format(inspect.currentframe().f_code.co_name, str(err)))
    #     raise err

def get_all_devices(server, dev_channel) -> list:
    '''Returns list of all device objects and their properties within a channel. Returned object is JSON list.
    
    INPUTS:
    "dev_channel" - channel the device object exists

    RETURNS:
    list - data for the devices requested

    EXCEPTIONS:
    KepHTTPError - If urllib provides an HTTPError
    KepURLError - If urllib provides an URLError
    '''
    r = server._config_get(server.url + channel._create_url(dev_channel) + _create_url())
    return r.payload

def auto_tag_gen(server, device_path, job_ttl = None) -> KepServiceResponse:
    '''Executes Auto Tag Generation function on devices that support the feature in Kepware
    
    INPUTS:
    "server" - instance of the "server" class
    
    "device_path" -  path identifying device. Standard Kepware address decimal notation string including the 
    device such as "channel1.device1"

    RETURNS:
    KepServiceReturn with the result of the service either Code 202 (Accepted) or 429 (Too Busy)

    EXCEPTIONS:
    KepHTTPError - If urllib provides an HTTPError
    KepURLError - If urllib provides an URLError
    '''
    
    path_obj = kepconfig.path_split(device_path)
    try:
        url = server.url +channel._create_url(path_obj['channel']) + _create_url(path_obj['device']) + ATG_URL
        job = server._kep_service_execute(url, job_ttl)
        return job
    except KeyError as err:
        err_msg = 'Error: No {} identified in {} | Function: {}'.format(err,'device_path', inspect.currentframe().f_code.co_name)
        raise KepError(err_msg)

def get_all_tags_tag_groups(server, device_path) -> dict:
    '''Returns the properties of all "tag" and "tag group" objects for as specific
    device in Kepware. Returned object is a dict of tag list and tag group list.

    The returned object resembles below, nested based on how many 
    levels the tag_group namespace has tags or tag_groups:

    Ex.
    {
        'tags': [tag1_dict, tag2_dict,...],
        'tag_groups':[
            {
                tag_group1_properties,
                'tags': [tag1_dict, tag2_dict,...]
                'tag_groups':[sub_group1, subgroup2,...]
            }, 
            {
                tag_group2_properties,
                'tags': [tag1_dict, tag2_dict,...]
                'tag_groups':[sub_group1, subgroup2,...]
            },...]
    } 

    INPUTS:
    "server" - instance of the "server" class
    
    "device_path" - path identifying device. Standard Kepware address decimal notation string including the 
    device such as "channel1.device1"

    RETURNS:
    dict - data for the tag structure requested at "device_path" location

    EXCEPTIONS:
    KepHTTPError - If urllib provides an HTTPError
    KepURLError - If urllib provides an URLError
    '''

    r = tag.get_full_tag_structure(server, device_path,recursive=True)
    return r

def get_device_structure(server, device_path) -> dict:
    '''Returns the properties of "device" and includes all "tag" and "tag group" objects for as specific
    device in Kepware. Returned object is a dict of device properties including a tag list and tag group list.

    The returned object resembles below, nested based on how many 
    levels the tag_group namespace has tags or tag_groups:

    Ex.
    {
        device_properties,
        'tags': [tag1_dict, tag2_dict,...],
        'tag_groups':[
            {
                tag_group1_properties,
                'tags': [tag1_dict, tag2_dict,...]
                'tag_groups':[sub_group1, subgroup2,...]
            }, 
            {
                tag_group2_properties,
                'tags': [tag1_dict, tag2_dict,...]
                'tag_groups':[sub_group1, subgroup2,...]
            },...]
    } 

    INPUTS:
    "server" - instance of the "server" class
    
    "device_path" - path identifying device. Standard Kepware address decimal notation string including the 
    device such as "channel1.device1"

    RETURNS:
    dict - data for the device structure requested at "device_path" location

    EXCEPTIONS:
    KepHTTPError - If urllib provides an HTTPError
    KepURLError - If urllib provides an URLError
    '''

    tags = tag.get_full_tag_structure(server, device_path,recursive=True)
    device_properties = get_device(server,device_path)
    return {**device_properties, **tags}



# -------------------------------------------------------------------------
# Copyright (c) PTC Inc. and/or all its affiliates. All rights reserved.
# See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

r"""`mapping` exposes an API to allow modifications (add, delete, modify) to 
column mapping objects in a Datalogger log group within the Kepware Configuration API
"""

from . import log_group as Log_Group
from ..error import KepError, KepHTTPError

MAPPING_ROOT = '/column_mappings'

def _create_url(mapping = None):
    '''Creates url object for the "column_mappings" branch of Kepware's project tree. Used 
    to build a part of Kepware Configuration API URL structure

    Returns the mapping specific url when a value is passed as the column_mapping name.
    '''

    if mapping == None:
        return '{}'.format(MAPPING_ROOT)
    else:
        return '{}/{}'.format(MAPPING_ROOT, mapping)

def modify_mapping(server, log_group, DATA, mapping = None, force = False) -> bool:
    '''Modify a column mapping object and it's properties in Kepware. If a "mapping" is not provided as an input,
    you need to identify the column mapping in the 'common.ALLTYPES_NAME' property field in the "DATA". It will 
    assume that is the column mapping that is to be modified.

    INPUTS:
    "server" - instance of the "server" class

    "log_group" - name of log group that mapping exists

    "DATA" - properly JSON object (dict) of the agent properties to be modified

    "mapping" (optional) - column mapping to modify in the log group. Only needed if not existing in "DATA"

    "force" (optional) - if True, will force the configuration update to the Kepware server

    RETURNS:
    True - If a "HTTP 200 - OK" is received from Kepware

    EXCEPTIONS:
    KepHTTPError - If urllib provides an HTTPError
    KepURLError - If urllib provides an URLError
    '''
    
    mapping_data = server._force_update_check(force, DATA)
    
    if mapping == None:
        try:
            r = server._config_update(server.url + Log_Group._create_url(log_group) + _create_url(mapping_data['common.ALLTYPES_NAME']), mapping_data)
            if r.code == 200: return True 
            else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)
        except KeyError as err:
            err_msg = 'Error: No column mapping identified in DATA | Key Error: {}'.format(err)
            raise KepError(err_msg)
        # except:
        #     return 'Error: Error with {}'.format(inspect.currentframe().f_code.co_name)
    else:
        r = server._config_update(server.url + Log_Group._create_url(log_group) + _create_url(mapping), mapping_data)
        if r.code == 200: return True 
        else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)

def get_mapping(server, log_group, mapping) -> dict:
    '''Returns the properties of the mapping object. Returned object is JSON.
    
    INPUTS:
    "server" - instance of the "server" class

    "log_group" - name of log group that mapping exists

    "mapping" - name of column mapping to retrieve properties for

    RETURNS:
    dict - data for the column mapping requested

    EXCEPTIONS:
    KepHTTPError - If urllib provides an HTTPError
    KepURLError - If urllib provides an URLError
    '''
    r = server._config_get(server.url + Log_Group._create_url(log_group) + _create_url(mapping))
    return r.payload

def get_all_mappings(server, log_group) -> list:
    '''Returns the properties of all column mapping objects for a log group. Returned object is JSON list.
    
    INPUTS:
    "server" - instance of the "server" class

    "log_group" - name of log group

    RETURNS:
    list - data for the column mappings requested

    EXCEPTIONS:
    KepHTTPError - If urllib provides an HTTPError
    KepURLError - If urllib provides an URLError
    '''
    r = server._config_get(server.url + Log_Group._create_url(log_group) + _create_url())
    return r.payload



# -------------------------------------------------------------------------
# Copyright (c), PTC Inc. and/or all its affiliates. All rights reserved.
# See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

r"""`ua_server` exposes an API to allow modifications (add, delete, modify) to 
OPC UA Server endpoints within the Kepware Administration through the Kepware Configuration API
"""
from typing import Union
from ..error import KepHTTPError, KepError


UA_ROOT = '/admin/ua_endpoints'

def _create_url(endpoint = None):
    '''Creates url object for the "server_users" branch of Kepware's project tree. Used 
    to build a part of Kepware Configuration API URL structure

    Returns the user specific url when a value is passed as the user name.
    '''
    
    if endpoint == None:
        return UA_ROOT
    else:
        return '{}/{}'.format(UA_ROOT,endpoint)

def add_endpoint(server, DATA) -> Union[bool, list]:
    '''Add an "endpoint" or multiple "endpoint" objects to Kepware UA Server by passing a 
    list of endpoints to be added all at once.

    INPUTS:
    "server" - instance of the "server" class

    "DATA" - properly JSON object (dict) of the endpoint
    expected by Kepware Configuration API

    RETURNS:
    True - If a "HTTP 201 - Created" is received from Kepware

    List - If a "HTTP 207 - Multi-Status" is received from Kepware with a list of dict error responses for all 
    endpoints added that failed.

    EXCEPTIONS:
    KepHTTPError - If urllib provides an HTTPError
    KepURLError - If urllib provides an URLError
    '''

    r = server._config_add(server.url + _create_url(), DATA)
    if r.code == 201: return True
    elif r.code == 207:
        errors = [] 
        for item in r.payload:
            if item['code'] != 201:
                errors.append(item)
        return errors
    else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)

def del_endpoint(server, endpoint) -> bool:
    '''Delete a "endpoint" object in Kepware UA Server
    
    INPUTS:
    "server" - instance of the "server" class

    "endpoint" - name of endpoint
    
    RETURNS:
    True - If a "HTTP 200 - OK" is received from Kepware

    EXCEPTIONS:
    KepHTTPError - If urllib provides an HTTPError
    KepURLError - If urllib provides an URLError
    '''

    r = server._config_del(server.url + _create_url(endpoint))
    if r.code == 200: return True 
    else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)

def modify_endpoint(server, DATA, endpoint = None) -> bool:
    '''Modify a endpoint object and it's properties in Kepware UA Server. If a "endpoint" is not provided as an input,
    you need to identify the endpoint in the 'common.ALLTYPES_NAME' property field in the "DATA". It will 
    assume that is the endpoint that is to be modified.

    INPUTS:
    "server" - instance of the "server" class

    "DATA" - properly JSON object (dict) of the endpoint properties to be modified.

    "endpoint" (optional) - name of endpoint to modify. Only needed if not existing in  "DATA"
    
    RETURNS:
    True - If a "HTTP 200 - OK" is received from Kepware

    EXCEPTIONS:
    KepHTTPError - If urllib provides an HTTPError
    KepURLError - If urllib provides an URLError
    '''
    
    # channel_data = server._force_update_check(force, DATA)
    if endpoint == None:
        try:
            r = server._config_update(server.url + _create_url(DATA['common.ALLTYPES_NAME']), DATA)
            if r.code == 200: return True 
            else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)
        except KeyError as err:
            err_msg = 'Error: No UA Endpoint identified in DATA | Key Error: {}'.format(err)
            raise KepError(err_msg)

        # except Exception as e:
        #     return 'Error: Error with {}: {}'.format(inspect.currentframe().f_code.co_name, str(e))
    else:
        r = server._config_update(server.url + _create_url(endpoint), DATA)
        if r.code == 200: return True 
        else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)

def get_endpoint(server, endpoint) -> dict:
    '''Returns the properties of the endpoint object. Returned object is JSON.
    
    INPUTS:
    "server" - instance of the "server" class

    "endpoint" - name of endpoint
    
    RETURNS:
    dict - data for the endpoint requested

    EXCEPTIONS:
    KepHTTPError - If urllib provides an HTTPError
    KepURLError - If urllib provides an URLError
    '''

    r = server._config_get(server.url + _create_url(endpoint))
    return r.payload

def get_all_endpoints(server) -> list:
    '''Returns list of all endpoint objects and their properties. Returned object is JSON list.
    
    INPUTS:
    "server" - instance of the "server" class
    
    RETURNS:
    list - data for all endpoints requested

    EXCEPTIONS:
    KepHTTPError - If urllib provides an HTTPError
    KepURLError - If urllib provides an URLError
    '''

    r = server._config_get(server.url + _create_url())
    return r.payload


# -------------------------------------------------------------------------
# Copyright (c) PTC Inc. and/or all its affiliates. All rights reserved.
# See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------


r"""`agent` exposes an API to allow modifications (add, delete, modify) to 
Iot Gateway agent objects within the Kepware Configuration API
"""

# from .. import connection 
from typing import Union
from .. import iot_gateway as IOT
from ..error import KepError, KepHTTPError
import inspect

IOT_ROOT_URL = '/project/_iot_gateway'
MQTT_CLIENT_URL = '/mqtt_clients'
REST_CLIENT_URL = '/rest_clients'
REST_SERVER_URL = '/rest_servers'
THINGWORX_URL = '/thingworx_clients'

def _create_url(agent_type, agent = None):
    '''Creates url object for the "agent" branch of Kepware's project tree. Used 
    to build a part of Kepware Configuration API URL structure

    Returns the agent specific url when a value is passed as the agent name.
    '''

    if agent == None:
        if agent_type == IOT.MQTT_CLIENT_AGENT:
            return '{}{}'.format(IOT_ROOT_URL, MQTT_CLIENT_URL)
        elif agent_type == IOT.REST_CLIENT_AGENT:
            return '{}{}'.format(IOT_ROOT_URL, REST_CLIENT_URL)
        elif agent_type == IOT.REST_SERVER_AGENT:
            return '{}{}'.format(IOT_ROOT_URL, REST_SERVER_URL)
        elif agent_type == IOT.THINGWORX_AGENT:
            return '{}{}'.format(IOT_ROOT_URL, THINGWORX_URL)
        else:
            pass
    else:
        if agent_type == IOT.MQTT_CLIENT_AGENT:
            return '{}{}/{}'.format(IOT_ROOT_URL, MQTT_CLIENT_URL, agent)
        elif agent_type == IOT.REST_CLIENT_AGENT:
            return '{}{}/{}'.format(IOT_ROOT_URL, REST_CLIENT_URL, agent)
        elif agent_type == IOT.REST_SERVER_AGENT:
            return '{}{}/{}'.format(IOT_ROOT_URL, REST_SERVER_URL,agent)
        elif agent_type == IOT.THINGWORX_AGENT:
            return '{}{}/{}'.format(IOT_ROOT_URL, THINGWORX_URL, agent)
        else:
            pass


def add_iot_agent(server, DATA, agent_type = None) -> Union[bool, list]:
    '''Add a  "agent" or multiple "agent" objects of a specific type to Kepware's IoT Gateway. Can be used to pass children of an
    agent object such as iot items. This allows you to create an agent and iot items if desired.

    Additionally it can be used to pass a list of agents and it's children to be added all at once.

    INPUTS:
    "server" - instance of the "server" class

    *DATA* - properly JSON object (dict) of the agent and it's children
    expected by Kepware Configuration API

    "agent_type" (optional) - agent type to add to IoT Gateway. Only needed if not existing in "DATA"

    RETURNS:
    True - If a "HTTP 201 - Created" is received from Kepware

    List - If a "HTTP 207 - Multi-Status" is received from Kepware with a list of dict error responses for all 
    iot agents added that failed.

    EXCEPTIONS:
    KepHTTPError - If urllib provides an HTTPError
    KepURLError - If urllib provides an URLError
    '''
    
    if agent_type == None:
        try:
            r = server._config_update(server.url + _create_url(DATA['iot_gateway.AGENTTYPES_TYPE']), DATA)
            if r.code == 201: return True 
            else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)
        except KeyError as err:
            err_msg = 'Error: No agent identified in DATA | Key Error: {}'.format(err)
            raise KepError(err_msg)
        # except:
        #     return 'Error: Error with {}'.format(inspect.currentframe().f_code.co_name)
    else:
        r = server._config_add(server.url + _create_url(agent_type), DATA)
        if r.code == 201: return True 
        elif r.code == 207:
            errors = [] 
            for item in r.payload:
                if item['code'] != 201:
                    errors.append(item)
            return errors
        else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)

def del_iot_agent(server, agent, agent_type) -> bool:
    '''Delete a "agent" object in Kepware. This will delete all children as well
    
    INPUTS:
    "server" - instance of the "server" class

    "agent" - name of IoT Agent

    "agent_type" - agent type to delete to IoT Gateway

    RETURNS:
    True - If a "HTTP 200 - OK" is received from Kepware

    EXCEPTIONS:
    KepHTTPError - If urllib provides an HTTPError
    KepURLError - If urllib provides an URLError
    '''
    r = server._config_del(server.url + _create_url(agent_type, agent))
    if r.code == 200: return True 
    else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)

def modify_iot_agent(server, DATA, agent = None, agent_type = None, force = False) -> bool:
    '''Modify a agent object and it's properties in Kepware. If a "agent" is not provided as an input,
    you need to identify the agent in the 'common.ALLTYPES_NAME' property field in the "DATA". It will 
    assume that is the agent that is to be modified.

    INPUTS:
    "server" - instance of the "server" class

    "DATA" - properly JSON object (dict) of the agent properties to be modified

    "agent" (optional) - name of IoT Agent. Only needed if not existing in "DATA"

    "agent_type" (optional) -agent type to modify to IoT Gateway. Only needed if not existing in "DATA"

    "force" (optional) - if True, will force the configuration update to the Kepware server

    RETURNS:
    True - If a "HTTP 200 - OK" is received from Kepware

    EXCEPTIONS:
    KepHTTPError - If urllib provides an HTTPError
    KepURLError - If urllib provides an URLError
    '''
    
    agent_data = server._force_update_check(force, DATA)
    
    if agent_type == None:
        if 'iot_gateway.AGENTTYPES_TYPE' in DATA:
            agent_type = DATA['iot_gateway.AGENTTYPES_TYPE']
        else:
            err_msg = 'Error: Error with {}: {}'.format(inspect.currentframe().f_code.co_name, 'No Agent type defined.')
            raise KepError(err_msg)
    if agent == None:
        try:
            r = server._config_update(server.url + _create_url(agent_type, agent_data['common.ALLTYPES_NAME']), agent_data)
            if r.code == 200: return True 
            else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)
        except KeyError as err:
            err_msg = 'Error: No agent identified in DATA | Key Error: {}'.format(err)
            raise KepError(err_msg)
        # except:
        #     return 'Error: Error with {}'.format(inspect.currentframe().f_code.co_name)
    else:
        r = server._config_update(server.url + _create_url(agent_type, agent), agent_data)
        if r.code == 200: return True 
        else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)

def get_iot_agent(server, agent, agent_type) -> dict:
    '''Returns the properties of the agent object. Returned object is JSON.
    
    INPUTS:
    "server" - instance of the "server" class

    "agent" - name of IoT Agent

    "agent_type" - agent type

    RETURNS:
    dict - data for the IoT Agent requested

    EXCEPTIONS:
    KepHTTPError - If urllib provides an HTTPError
    KepURLError - If urllib provides an URLError
    '''
    r = server._config_get(server.url + _create_url(agent_type, agent))
    return r.payload

def get_all_iot_agents(server, agent_type) -> list:
    '''Returns the properties of all agent objects for a specific agent type. Returned object is JSON list.
    
    INPUTS:
    "server" - instance of the "server" class

    "agent_type" - agent type

    RETURNS:
    list - data for the IoT Agents requested

    EXCEPTIONS:
    KepHTTPError - If urllib provides an HTTPError
    KepURLError - If urllib provides an URLError
    '''
    r = server._config_get(server.url + _create_url(agent_type))
    return r.payload



# -------------------------------------------------------------------------
# Copyright (c) 2020, PTC Inc. and/or all its affiliates. All rights reserved.
# See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------


__path__ = __import__("pkgutil").extend_path(__path__, __name__)
from . import agent, iot_items

#IoT Gateway Agent Types
MQTT_CLIENT_AGENT = 'MQTT Client'
REST_CLIENT_AGENT = 'REST Client'
REST_SERVER_AGENT = 'REST Server'
THINGWORX_AGENT = 'ThingWorx'



# -------------------------------------------------------------------------
# Copyright (c) PTC Inc. and/or all its affiliates. All rights reserved.
# See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------


r"""`ranges` exposes an API to allow modifications (add, delete, modify) to 
range objects in exchanges for EGD devices within the Kepware Configuration API
"""

from typing import Union
from .. import egd as EGD

RANGES_ROOT = '/ranges'

def _create_url(device_path, ex_type, exchange_name, range = None):
    '''Creates url object for the "range" branch of Kepware's project tree. Used 
    to build a part of Kepware Configuration API URL structure

    Returns the range specific url when a value is passed as the range name.
    '''
    exchange_root = EGD.exchange._create_url(device_path, ex_type, exchange_name)

    if range == None:
        return '{}{}'.format(exchange_root, RANGES_ROOT)
    else:
        return '{}{}/{}'.format(exchange_root, RANGES_ROOT, range)

def add_range(server, device_path, ex_type, exchange_name, DATA) -> Union[bool, list]:
    '''Add a "range" or multiple "range" objects to Kepware. This allows you to 
    create a range or multiple ranges all in one function, if desired.

    When passing multiple ranges, they will be populated in the same order
    in the list sent. Ensure you provide the list in the order desired.

    INPUTS:

    "server" - instance of the "server" class

    "device_path" - path to exchanges and their ranges. Standard Kepware address decimal 
    notation string such as "channel1.device1"

    "ex_type" - type of exchange either consumer or producer

    "exchange_name" - name of exchange to add range to

    "DATA" - properly JSON object (dict) of the range or ranges

    RETURNS:

    True - If a "HTTP 201 - Created" is received from Kepware
    
    List - If a "HTTP 207 - Multi-Status" is received from Kepware with a list of dict error responses for all 
    ranges added that failed.

    False - If a non-expected "2xx successful" code is returned

    EXCEPTIONS:

    KepHTTPError - If urllib provides an HTTPError
    KepURLError - If urllib provides an URLError
    '''

    r = server._config_add(server.url + _create_url(device_path, ex_type, exchange_name), DATA)
    if r.code == 201: return True
    elif r.code == 207:
        errors = [] 
        for item in r.payload:
            if item['code'] != 201:
                errors.append(item)
        return errors
    else: return False

def del_range(server, device_path, ex_type, exchange_name, range_name) -> bool:
    '''Delete a "range" object in Kepware.
    
    INPUTS:

    "server" - instance of the "server" class

    "device_path" - path to exchanges and their ranges. Standard Kepware address decimal 
    notation string such as "channel1.device1"

    "ex_type" - type of exchange either consumer or producer

    "exchange_name" - name of exchange that range is located

    "range_name" - name of range
    
    RETURNS:

    True - If a "HTTP 200 - OK" is received from Kepware

    EXCEPTIONS:

    KepHTTPError - If urllib provides an HTTPError
    KepURLError - If urllib provides an URLError
    '''

    r = server._config_del(server.url + _create_url(device_path, ex_type, exchange_name, range_name))
    if r.code == 200: return True 
    else: return False

def modify_range(server, device_path, ex_type, exchange_name, DATA, range_name = None, force = False) -> bool:
    '''Modify a range object and it's properties in Kepware. If a "range_name" is not provided as an input,
    you need to identify the range in the 'common.ALLTYPES_NAME' property field in the "DATA". It will 
    assume that is the range that is to be modified.

    INPUTS:

    "server" - instance of the "server" class

    "device_path" - path to exchanges and their ranges. Standard Kepware address decimal 
    notation string such as "channel1.device1"

    "ex_type" - type of exchange either consumer or producer

    "exchange_name" - name of exchange that range is located

    "DATA" - properly JSON object (dict) of the range properties to be modified.

    "range_name" (optional) - name of range to modify. Only needed if not existing in  "DATA"

    "force" (optional) - if True, will force the configuration update to the Kepware server
    
    RETURNS:
    
    True - If a "HTTP 200 - OK" is received from Kepware

    EXCEPTIONS:
    
    KepHTTPError - If urllib provides an HTTPError
    KepURLError - If urllib provides an URLError
    '''
    
    range_data = server._force_update_check(force, DATA)
    if range_name == None:
        try:
            r = server._config_update(server.url + _create_url(device_path, ex_type, exchange_name, range_data['common.ALLTYPES_NAME']), range_data)
            if r.code == 200: return True 
            else: return False
        except KeyError as err:
            print('Error: No range identified in DATA | Key Error: {}'.format(err))
            return False
        # except Exception as e:
        #     return 'Error: Error with {}: {}'.format(inspect.currentframe().f_code.co_name, str(e))
    else:
        r = server._config_update(server.url + _create_url(device_path, ex_type, exchange_name, range_name), range_data)
        if r.code == 200: return True 
        else: return False

def get_range(server, device_path, ex_type, exchange_name, range_name = None) -> Union[dict, list]:
    '''Returns the properties of the range object or a list of all ranges. Returned object is JSON.
    
    INPUTS:

    "server" - instance of the "server" class

    "device_path" - path to exchanges and their ranges. Standard Kepware address decimal 
    notation string such as "channel1.device1"

    "ex_type" - type of exchange either consumer or producer

    "exchange_name" - name of exchange that range is located

    "range_name" - name of range
    
    RETURNS:

    JSON - data for the range requested or a list of ranges and their properties

    EXCEPTIONS:

    KepHTTPError - If urllib provides an HTTPError
    KepURLError - If urllib provides an URLError
    '''
    if range_name == None:
        r = server._config_get(server.url + _create_url(device_path, ex_type, exchange_name))
    else:
        r = server._config_get(server.url + _create_url(device_path, ex_type, exchange_name, range_name))
    return r.payload


# -------------------------------------------------------------------------
# Copyright (c) PTC Inc. and/or all its affiliates. All rights reserved.
# See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------


r"""`channel` exposes an API to allow modifications (add, delete, modify) to 
channel objects within the Kepware Configuration API
"""

 
import inspect
from ..error import KepHTTPError, KepError
from typing import Union
from. import device

CHANNEL_ROOT = '/project/channels'

def _create_url(channel = None):
    '''Creates url object for the "channel" branch of Kepware's project tree. Used 
    to build a part of Kepware Configuration API URL structure

    Returns the channel specific url when a value is passed as the channel name.
    '''
    
    if channel == None:
        return CHANNEL_ROOT
    else:
        return '{}/{}'.format(CHANNEL_ROOT,channel)

def add_channel(server, DATA) -> Union[bool, list]:
    '''Add a "channel" or multiple "channel" objects to Kepware. Can be used to pass children of a channel object 
    such as devices and tags/tag groups. This allows you to create a channel, it's devices and tags 
    all in one function, if desired.

    Additionally it can be used to pass a list of channels and it's children to be added all at once.

    INPUTS:
    "server" - instance of the "server" class

    "DATA" - properly JSON object (dict) of the channel and it's children
    expected by Kepware Configuration API

    RETURNS:
    True - If a "HTTP 201 - Created" is received from Kepware
    
    List - If a "HTTP 207 - Multi-Status" is received from Kepware with a list of dict error responses for all 
    channels added that failed.
        

    EXCEPTIONS:
    KepHTTPError - If urllib provides an HTTPError
    KepURLError - If urllib provides an URLError
    '''

    r = server._config_add(server.url + _create_url(), DATA)
    if r.code == 201: return True
    elif r.code == 207:
        errors = [] 
        for item in r.payload:
            if item['code'] != 201:
                errors.append(item)
        return errors
    else: 
        raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)

def del_channel(server, channel) -> bool:
    '''Delete a "channel" object in Kepware. This will delete all children as well
    
    INPUTS:
    "server" - instance of the "server" class

    "channel" - name of channel
    
    RETURNS:
    True - If a "HTTP 200 - OK" is received from Kepware

    EXCEPTIONS:
    KepHTTPError - If urllib provides an HTTPError
    KepURLError - If urllib provides an URLError
    '''

    r = server._config_del(server.url + _create_url(channel))
    if r.code == 200: return True 
    else: 
        raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)

def modify_channel(server, DATA, channel = None, force = False) -> bool:
    '''Modify a channel object and it's properties in Kepware. If a "channel" is not provided as an input,
    you need to identify the channel in the 'common.ALLTYPES_NAME' property field in the "DATA". It will 
    assume that is the channel that is to be modified.

    INPUTS:
    "server" - instance of the "server" class

    "DATA" - properly JSON object (dict) of the channel properties to be modified.

    "channel" (optional) - name of channel to modify. Only needed if not existing in  "DATA"

    "force" (optional) - if True, will force the configuration update to the Kepware server
    
    RETURNS:
    True - If a "HTTP 200 - OK" is received from Kepware

    EXCEPTIONS:
    KepHTTPError - If urllib provides an HTTPError
    KepURLError - If urllib provides an URLError
    '''
    
    channel_data = server._force_update_check(force, DATA)
    if channel == None:
        try:
            r = server._config_update(server.url + _create_url(channel_data['common.ALLTYPES_NAME']), channel_data)
            if r.code == 200: return True 
            else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)
        except KeyError as err:
            err_msg = 'Error: No Channel identified in DATA | Key Error: {}'.format(err)
            raise KepError(err_msg)
        # except Exception as e:
        #     return 'Error: Error with {}: {}'.format(inspect.currentframe().f_code.co_name, str(e))
    else:
        r = server._config_update(server.url + _create_url(channel), channel_data)
        if r.code == 200: return True 
        else: 
            raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)

def get_channel(server, channel)  -> dict:
    '''Returns the properties of the channel object. Returned object is JSON.
    
    INPUTS:
    "server" - instance of the "server" class

    "channel" - name of channel
    
    RETURNS:
    dict - data for the channel requested

    EXCEPTIONS:
    KepHTTPError - If urllib provides an HTTPError
    KepURLError - If urllib provides an URLError
    '''

    r = server._config_get(server.url + _create_url(channel))
    return r.payload

def get_all_channels(server) -> list:
    '''Returns list of all channel objects and their properties. Returned object is JSON list.
    
    INPUTS:
    "server" - instance of the "server" class
    
    RETURNS:
    list - data for the channel requested

    EXCEPTIONS:
    KepHTTPError - If urllib provides an HTTPError
    KepURLError - If urllib provides an URLError
    '''

    r = server._config_get(server.url + _create_url())
    return r.payload

def get_channel_structure(server, channel) -> dict:
    '''Returns the properties of "channel" and includes all "devices" and the "tag" and "tag group" objects for a 
    channel in Kepware. Returned object is a dict of channel properties including a device list with 
    tag lists and tag group lists.

    The returned object resembles below, nested based on how many 
    levels the tag_group namespace has tags or tag_groups:

    Ex.
    {
        channel_properties,
        'devices: [
            {
                device1_properties,
                'tags': [tag1_dict, tag2_dict,...],
                'tag_groups':[
                    {
                        tag_group1_properties,
                        'tags': [tag1_dict, tag2_dict,...]
                        'tag_groups':[sub_group1, subgroup2,...]
                    }, 
                    {
                        tag_group2_properties,
                        'tags': [tag1_dict, tag2_dict,...]
                        'tag_groups':[sub_group1, subgroup2,...]
                    },...]
            },...]
    }       

    INPUTS:
    "server" - instance of the "server" class
    
    "channel" - name of channel

    RETURNS:
    dict - data for the device structure requested for "channel"

    EXCEPTIONS:
    KepHTTPError - If urllib provides an HTTPError
    KepURLError - If urllib provides an URLError
    '''

    channel_properties = get_channel(server, channel)
    device_list = device.get_all_devices(server,channel)
    for dev in device_list:
        device_properties = []
        dev_struct = device.get_device_structure(server,channel + '.' + dev['common.ALLTYPES_NAME'])
        device_properties.append(dev_struct)
    return {**channel_properties,'device': device_properties}


# -------------------------------------------------------------------------
# Copyright (c) PTC Inc. and/or all its affiliates. All rights reserved.
# See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

r"""`triggers` exposes an API to allow modifications (add, delete, modify) to 
trigger objects in a Datalogger log group within the Kepware Configuration API
"""

from typing import Union
from . import log_group as Log_Group
from ..error import KepError, KepHTTPError

TRIGGERS_ROOT = '/triggers'

def _create_url(trigger = None):
    '''Creates url object for the "trigger" branch of Kepware's project tree. Used 
    to build a part of Kepware Configuration API URL structure

    Returns the trigger specific url when a value is passed as the trigger name.
    '''

    if trigger == None:
        return '{}'.format(TRIGGERS_ROOT)
    else:
        return '{}/{}'.format(TRIGGERS_ROOT, trigger)


def add_trigger(server, log_group, DATA) -> Union[bool, list]:
    '''Add a "trigger" or multiple "trigger" objects to a log group in Kepware's Datalogger. It can 
    be used to pass a list of triggers to be added all at once.

    INPUTS:
    "server" - instance of the "server" class

    "log_group" - name of log group that the triggers will be added

    *DATA* - properly JSON object (dict) of the trigger expected by Kepware Configuration API

    RETURNS:
    True - If a "HTTP 201 - Created" is received from Kepware

    List - If a "HTTP 207 - Multi-Status" is received from Kepware with a list of dict error responses for all 
    triggers added that failed.

    EXCEPTIONS:
    KepHTTPError - If urllib provides an HTTPError
    KepURLError - If urllib provides an URLError
    '''

    r = server._config_add(server.url + Log_Group._create_url(log_group) + _create_url(), DATA)
    if r.code == 201: return True
    elif r.code == 207:
        errors = [] 
        for item in r.payload:
            if item['code'] != 201:
                errors.append(item)
        return errors 
    else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)

def del_trigger(server, log_group, trigger) -> bool:
    '''Delete a "trigger" object of a log group in Kepware's Datalogger.
    
    INPUTS:
    "server" - instance of the "server" class

    "log_group" - name of log group that trigger exists

    "trigger" - name of trigger to delete

    RETURNS:
    True - If a "HTTP 200 - OK" is received from Kepware

    EXCEPTIONS:
    KepHTTPError - If urllib provides an HTTPError
    KepURLError - If urllib provides an URLError
    '''
    r = server._config_del(server.url + Log_Group._create_url(log_group) + _create_url(trigger))
    if r.code == 200: return True
    else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)

def modify_trigger(server, log_group, DATA, trigger = None, force = False)  -> bool:
    '''Modify a trigger object and it's properties in Kepware. If a "trigger" is not provided as an input,
    you need to identify the trigger in the 'common.ALLTYPES_NAME' property field in the "DATA". It will 
    assume that is the trigger that is to be modified.

    INPUTS:
    "server" - instance of the "server" class

    "log_group" - name of log group that trigger exists

    "DATA" - properly JSON object (dict) of the agent properties to be modified

    "trigger" (optional) - trigger to modify in the log group. Only needed if not existing in "DATA"

    "force" (optional) - if True, will force the configuration update to the Kepware server

    RETURNS:
    True - If a "HTTP 200 - OK" is received from Kepware

    EXCEPTIONS:
    KepHTTPError - If urllib provides an HTTPError
    KepURLError - If urllib provides an URLError
    '''
    
    trigger_data = server._force_update_check(force, DATA)
    
    if trigger == None:
        try:
            r = server._config_update(server.url + Log_Group._create_url(log_group) + _create_url(trigger_data['common.ALLTYPES_NAME']), trigger_data)
            if r.code == 200: return True 
            else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)
        except KeyError as err:
            err_msg = 'Error: No trigger identified in DATA | Key Error: {}'.format(err)
            raise KepError(err_msg)
        # except:
        #     return 'Error: Error with {}'.format(inspect.currentframe().f_code.co_name)
    else:
        r = server._config_update(server.url + Log_Group._create_url(log_group) + _create_url(trigger), trigger_data)
        if r.code == 200: return True 
        else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)

def get_trigger(server, log_group, trigger) -> dict:
    '''Returns the properties of the trigger object. Returned object is JSON.
    
    INPUTS:
    "server" - instance of the "server" class

    "log_group" - name of log group that trigger exists

    "trigger" - name of trigger to retrieve properties for

    RETURNS:
    dict - data for the trigger requested

    EXCEPTIONS:
    KepHTTPError - If urllib provides an HTTPError
    KepURLError - If urllib provides an URLError
    '''
    r = server._config_get(server.url + Log_Group._create_url(log_group) + _create_url(trigger))
    return r.payload

def get_all_triggers(server, log_group) -> list:
    '''Returns the properties of all trigger objects for a log group. Returned object is JSON list.
    
    INPUTS:
    "server" - instance of the "server" class

    "log_group" - name of log group

    RETURNS:
    list - data for the triggers requested

    EXCEPTIONS:
    KepHTTPError - If urllib provides an HTTPError
    KepURLError - If urllib provides an URLError
    '''
    r = server._config_get(server.url + Log_Group._create_url(log_group) + _create_url())
    return r.payload



# -------------------------------------------------------------------------
# Copyright (c) PTC Inc. and/or all its affiliates. All rights reserved.
# See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

r"""`log_group` exposes an API to allow modifications (add, delete, modify) to 
log group objects in DataLogger within the Kepware Configuration API
"""
from typing import Union
from ..connection import KepServiceResponse
from ..error import KepError, KepHTTPError

ENABLE_PROPERTY = 'datalogger.LOG_GROUP_ENABLED'
LOG_GROUP_ROOT = '/project/_datalogger/log_groups'
SERVICES_ROOT = '/services'
def _create_url(log_group = None):
    '''Creates url object for the "log_group" branch of Kepware's project tree. Used 
    to build a part of Kepware Configuration API URL structure

    Returns the agent specific url when a value is passed as the agent name.
    '''

    if log_group == None:
        return '{}'.format(LOG_GROUP_ROOT)
    else:
        return '{}/{}'.format(LOG_GROUP_ROOT, log_group)


def add_log_group(server, DATA) -> Union[bool, list]:
    '''Add a "log group" or multiple "log groups" objects to Kepware's DataLogger. It can be used 
    to pass a list of log groups to be added all at once.

    INPUTS:
    "server" - instance of the "server" class

    *DATA* - properly JSON object (dict) of the log group expected by Kepware Configuration API

    RETURNS:
    True - If a "HTTP 201 - Created" is received from Kepware

    List - If a "HTTP 207 - Multi-Status" is received from Kepware with a list of dict error responses for all 
    log groups added that failed.

    EXCEPTIONS:
    KepHTTPError - If urllib provides an HTTPError
    KepURLError - If urllib provides an URLError
    '''

    r = server._config_add(server.url + _create_url(), DATA)
    if r.code == 201: return True 
    elif r.code == 207:
        errors = [] 
        for item in r.payload:
            if item['code'] != 201:
                errors.append(item)
        return errors
    else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)

def del_log_group(server, log_group) -> bool:
    '''Delete a "log group" object in Kepware's Datalogger.
    
    INPUTS:
    "server" - instance of the "server" class

    "log_group" - name of log group

    RETURNS:
    True - If a "HTTP 200 - OK" is received from Kepware

    EXCEPTIONS:
    KepHTTPError - If urllib provides an HTTPError
    KepURLError - If urllib provides an URLError
    '''
    r = server._config_del(server.url + _create_url(log_group))
    if r.code == 200: return True 
    else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)

def modify_log_group(server, DATA, log_group = None, force = False) -> bool:
    '''Modify a log group object and it's properties in Kepware's Datalogger. If a "log group" is not provided as an input,
    you need to identify the log group in the 'common.ALLTYPES_NAME' property field in the "DATA". It will 
    assume that is the log group that is to be modified.

    INPUTS:
    "server" - instance of the "server" class

    "DATA" - properly JSON object (dict) of the agent properties to be modified

    "log_group" (optional) - name of log group. Only needed if not existing in "DATA"

    "force" (optional) - if True, will force the configuration update to the Kepware server

    RETURNS:
    True - If a "HTTP 200 - OK" is received from Kepware

    EXCEPTIONS:
    KepHTTPError - If urllib provides an HTTPError
    KepURLError - If urllib provides an URLError
    '''
    
    log_group_data = server._force_update_check(force, DATA)
    
    if log_group == None:
        try:
            r = server._config_update(server.url + _create_url(log_group_data['common.ALLTYPES_NAME']), log_group_data)
            if r.code == 200: return True 
            else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)
        except KeyError as err:
            err_msg = 'Error: No log group identified in DATA | Key Error: {}'.format(err)
            raise KepError(err_msg)
        # except:
        #     return 'Error: Error with {}'.format(inspect.currentframe().f_code.co_name)
    else:
        r = server._config_update(server.url + _create_url(log_group), log_group_data)
        if r.code == 200: return True 
        else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)

def get_log_group(server, log_group) -> dict:
    '''Returns the properties of the log group object. Returned object is JSON.
    
    INPUTS:
    "server" - instance of the "server" class

    "log_group" - name of log group

    RETURNS:
    dict - data for the log group requested

    EXCEPTIONS:
    KepHTTPError - If urllib provides an HTTPError
    KepURLError - If urllib provides an URLError
    '''
    r = server._config_get(server.url + _create_url(log_group))
    return r.payload

def get_all_log_groups(server) -> list:
    '''Returns the properties of all log group objects for Kepware's Datalogger. Returned object is JSON list.
    
    INPUTS:
    "server" - instance of the "server" class

    RETURNS:
    list - data for the log groups requested

    EXCEPTIONS:
    KepHTTPError - If urllib provides an HTTPError
    KepURLError - If urllib provides an URLError
    '''
    r = server._config_get(server.url + _create_url())
    return r.payload

def enable_log_group(server, log_group) -> bool:
    '''Enable the log group. Returned object is JSON.
    
    INPUTS:
    "server" - instance of the "server" class

    "log_group" - name of log group

    RETURNS:
    True - If a "HTTP 200 - OK" is received from Kepware

    EXCEPTIONS:
    KepHTTPError - If urllib provides an HTTPError
    KepURLError - If urllib provides an URLError
    '''
    DATA = {ENABLE_PROPERTY: True}
    return modify_log_group(server, DATA, log_group)

def disable_log_group(server, log_group) -> bool:
    '''Disable the log group. Returned object is JSON.
    
    INPUTS:
    "server" - instance of the "server" class

    "log_group" - name of log group

    RETURNS:
    True - If a "HTTP 200 - OK" is received from Kepware

    EXCEPTIONS:
    KepHTTPError - If urllib provides an HTTPError
    KepURLError - If urllib provides an URLError
    '''
    DATA = {ENABLE_PROPERTY: False}
    return modify_log_group(server, DATA, log_group)

def reset_column_mapping_service(server, log_group, job_ttl = None) -> KepServiceResponse:
    '''Executes a ResetColumnMapping serivce call to the log group

    INPUTS:
    "server" - instance of the "server" class

    "log_group" - name of log group

    RETURNS:
    KepServiceResponse instance with job information
    
    EXCEPTIONS (If not HTTP 200 or 429 returned):
    KepHTTPError - If urllib provides an HTTPError
    KepURLError - If urllib provides an URLError
    '''

    url = server.url + _create_url(log_group) + SERVICES_ROOT + '/ResetColumnMapping'
    job = server._kep_service_execute(url, job_ttl)
    return job


# -------------------------------------------------------------------------
# Copyright (c) PTC Inc. and/or all its affiliates. All rights reserved.
# See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

r"""`log_items` exposes an API to allow modifications (add, delete, modify) to 
log item (tag) objects in a Datalogger log group within the Kepware Configuration API
"""
from typing import Union
from . import log_group as Log_Group
from ..error import KepError, KepHTTPError

LOG_ITEMS_ROOT = '/log_items'

def _create_url(log_item = None):
    '''Creates url object for the "log_item" branch of Kepware's project tree. Used 
    to build a part of Kepware Configuration API URL structure

    Returns the log_item specific url when a value is passed as the log_item name.
    '''

    if log_item == None:
        return '{}'.format(LOG_ITEMS_ROOT)
    else:
        return '{}/{}'.format(LOG_ITEMS_ROOT, log_item)


def add_log_item(server, log_group, DATA) -> Union[bool, list]:
    '''Add a "log item" or multiple "log item" objects to a log group in Kepware's Datalogger. It can 
    be used to pass a list of log items to be added all at once.

    INPUTS:
    "server" - instance of the "server" class

    "log_group" - name of log group that the log items will be added

    *DATA* - properly JSON object (dict) of the log item expected by Kepware Configuration API

    RETURNS:
    True - If a "HTTP 201 - Created" is received from Kepware

    List - If a "HTTP 207 - Multi-Status" is received from Kepware with a list of dict error responses for all 
    log items added that failed.

    EXCEPTIONS:
    KepHTTPError - If urllib provides an HTTPError
    KepURLError - If urllib provides an URLError
    '''

    r = server._config_add(server.url + Log_Group._create_url(log_group) + _create_url(), DATA)
    if r.code == 201: return True
    elif r.code == 207:
        errors = [] 
        for item in r.payload:
            if item['code'] != 201:
                errors.append(item)
        return errors 
    else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)

def del_log_item(server, log_group, log_item) -> bool:
    '''Delete a "log item" object of a log group in Kepware's Datalogger.
    
    INPUTS:
    "server" - instance of the "server" class

    "log_group" - name of log group that log item exists

    "log_item" - name of log item to delete

    RETURNS:
    True - If a "HTTP 200 - OK" is received from Kepware

    EXCEPTIONS:
    KepHTTPError - If urllib provides an HTTPError
    KepURLError - If urllib provides an URLError
    '''
    r = server._config_del(server.url + Log_Group._create_url(log_group) + _create_url(log_item))
    if r.code == 200: return True 
    else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)

def modify_log_item(server, log_group, DATA, log_item = None, force = False) -> bool:
    '''Modify a log_item object and it's properties in Kepware. If a "log_item" is not provided as an input,
    you need to identify the log_item in the 'common.ALLTYPES_NAME' property field in the "DATA". It will 
    assume that is the log_item that is to be modified.

    INPUTS:
    "server" - instance of the "server" class

    "log_group" - name of log group that log item exists

    "DATA" - properly JSON object (dict) of the agent properties to be modified

    "log_item" (optional) - log item to modify in the log group. Only needed if not existing in "DATA"

    "force" (optional) - if True, will force the configuration update to the Kepware server

    RETURNS:
    True - If a "HTTP 200 - OK" is received from Kepware

    EXCEPTIONS:
    KepHTTPError - If urllib provides an HTTPError
    KepURLError - If urllib provides an URLError
    '''
    
    log_item_data = server._force_update_check(force, DATA)
    
    if log_item == None:
        try:
            r = server._config_update(server.url + Log_Group._create_url(log_group) + _create_url(log_item_data['common.ALLTYPES_NAME']), log_item_data)
            if r.code == 200: return True 
            else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)
        except KeyError as err:
            err_msg ='Error: No log item identified in DATA | Key Error: {}'.format(err)
            raise KepError(err_msg)
        # except:
        #     return 'Error: Error with {}'.format(inspect.currentframe().f_code.co_name)
    else:
        r = server._config_update(server.url + Log_Group._create_url(log_group) + _create_url(log_item), log_item_data)
        if r.code == 200: return True 
        else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)

def get_log_item(server, log_group, log_item) -> dict:
    '''Returns the properties of the log item object. Returned object is JSON.
    
    INPUTS:
    "server" - instance of the "server" class

    "log_group" - name of log group that log item exists

    "log_item" - name of log item to retrieve properties for

    RETURNS:
    dict - data for the log item requested

    EXCEPTIONS:
    KepHTTPError - If urllib provides an HTTPError
    KepURLError - If urllib provides an URLError
    '''
    r = server._config_get(server.url + Log_Group._create_url(log_group) + _create_url(log_item))
    return r.payload

def get_all_log_items(server, log_group) -> list:
    '''Returns the properties of all log item objects for a log group. Returned object is JSON list.
    
    INPUTS:
    "server" - instance of the "server" class

    "log_group" - name of log group

    RETURNS:
    list - data for the log items requested

    EXCEPTIONS:
    KepHTTPError - If urllib provides an HTTPError
    KepURLError - If urllib provides an URLError
    '''
    r = server._config_get(server.url + Log_Group._create_url(log_group) + _create_url())
    return r.payload



# -------------------------------------------------------------------------
# Copyright (c) PTC Inc. and/or all its affiliates. All rights reserved.
# See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------


r"""`iot_items` exposes an API to allow modifications (add, delete, modify) to 
iot_items objects within the Kepware Configuration API
"""

from typing import Union
import kepconfig as helper
from .. import iot_gateway as IOT
from ..error import KepError, KepHTTPError

IOT_ITEMS_ROOT = '/iot_items'

def _create_url(tag = None):
    '''Creates url object for the "iot items" branch of Kepware's IoT Agents property model. Used 
    to build a part of Kepware Configuration API URL structure
    
    Returns the device specific url when a value is passed as the iot item name.
    '''
    if tag == None:
        return IOT_ITEMS_ROOT
    else: 
        normalized_tag = helper._address_dedecimal(tag)
        return '{}/{}'.format(IOT_ITEMS_ROOT,normalized_tag)


def add_iot_item(server, DATA, agent, agent_type) -> Union[bool, list]:
    '''Add a "iot item" or multiple "iot item" objects to Kepware's IoT Gateway agent. Additionally 
    it can be used to pass a list of iot items to be added to an agent all at once.

    INPUTS:
    "server" - instance of the "server" class

    "DATA" - properly JSON object (dict) of the iot item or list of items
    expected by Kepware Configuration API.

    "agent" - name of IoT Agent

    "agent_type" - agent type

    RETURNS:
    True - If a "HTTP 201 - Created" is received from Kepware

    List - If a "HTTP 207 - Multi-Status" is received from Kepware with a list of dict error responses for all 
    iot items added that failed.

    EXCEPTIONS:
    KepHTTPError - If urllib provides an HTTPError
    KepURLError - If urllib provides an URLError
    '''
    r = server._config_add(server.url + IOT.agent._create_url(agent_type, agent) + _create_url(), DATA)
    if r.code == 201: return True
    elif r.code == 207:
            errors = [] 
            for item in r.payload:
                if item['code'] != 201:
                    errors.append(item)
            return errors 
    else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)

def del_iot_item(server, iot_item, agent, agent_type) -> bool:
    '''Delete a "iot item" object in Kepware.

    INPUTS:
    "server" - instance of the "server" class

    "iot_item" - IoT item to delete

    "agent" - name of IoT Agent

    "agent_type" - agent type 

    RETURNS:
    True - If a "HTTP 200 - OK" is received from Kepware

    EXCEPTIONS:
    KepHTTPError - If urllib provides an HTTPError
    KepURLError - If urllib provides an URLError
    '''
    r = server._config_del(server.url + IOT.agent._create_url(agent_type, agent) + _create_url(iot_item))
    if r.code == 200: return True 
    else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)

def modify_iot_item(server, DATA, agent, agent_type, iot_item = None, force = False) -> bool:
    '''Modify a iot item object and it's properties in Kepware. If a "iot item" is not provided as an input,
    you need to identify the iot item in the 'common.ALLTYPES_NAME' property field in the "DATA". It will 
    assume that is the iot item that is to be modified.

    INPUTS:
    "server" - instance of the "server" class

    "DATA" - properly JSON object (dict) of the iot item properties to be modified.

    "agent" - name of IoT Agent

    "agent_type" - agent type 

    "iot_item" (optional) - IoT item to modify

    "force" (optional) - if True, will force the configuration update to the Kepware server

    RETURNS:
    True - If a "HTTP 200 - OK" is received from Kepware

    EXCEPTIONS:
    KepHTTPError - If urllib provides an HTTPError
    KepURLError - If urllib provides an URLError
    '''
    
    agent_data = server._force_update_check(force, DATA)
    if iot_item == None:
        try:
            r = server._config_update(server.url + IOT.agent._create_url(agent_type, agent) + _create_url(agent_data['common.ALLTYPES_NAME']), agent_data)
            if r.code == 200: return True 
            else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)
        except KeyError as err:
            err_msg = 'Error: No agent identified in DATA | Key Error: {}'.format(err)
            raise KepError(err_msg)
        # except:
        #     return 'Error: Error with {}'.format(inspect.currentframe().f_code.co_name)
    else:
        r = server._config_update(server.url + IOT.agent._create_url(agent_type, agent) + _create_url(iot_item), agent_data)
        if r.code == 200: return True 
        else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)

def get_iot_item(server, iot_item, agent, agent_type)-> dict:
    '''Returns the properties of the agent object. Returned object is JSON.

    INPUTS:
    "server" - instance of the "server" class
    
    "iot_item" - IoT item

    "agent" - name of IoT Agent

    "agent_type" - agent type

    RETURNS:
    dict - data for the IoT item requested

    EXCEPTIONS:
    KepHTTPError - If urllib provides an HTTPError
    KepURLError - If urllib provides an URLError
    '''
    r = server._config_get(server.url + IOT.agent._create_url(agent_type, agent) + _create_url(iot_item))
    return r.payload

def get_all_iot_items(server, agent, agent_type) -> list:
    '''Returns the properties of all iot item objects for an agent. Returned object is JSON list.
    
    INPUTS:
    "server" - instance of the "server" class

    "agent" - name of IoT Agent

    "agent_type" - agent type

    RETURNS:
    list - data for the IoT item requested

    EXCEPTIONS:
    KepHTTPError - If urllib provides an HTTPError
    KepURLError - If urllib provides an URLError
    '''
    r = server._config_get(server.url + IOT.agent._create_url(agent_type, agent) + _create_url())
    return r.payload


# -------------------------------------------------------------------------
# Copyright (c), PTC Inc. and/or all its affiliates. All rights reserved.
# See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

r"""`user_groups` exposes an API to allow modifications (add, delete, modify) to 
user_groups within the Kepware Administration User Manager through the Kepware Configuration API
"""
from typing import Union
from ..error import KepHTTPError, KepError


USERGROUPS_ROOT = '/admin/server_usergroups'
ENABLE_PROPERTY = 'libadminsettings.USERMANAGER_GROUP_ENABLED'

def _create_url(user_group = None):
    '''Creates url object for the "server_usergroups" branch of Kepware's project tree. Used 
    to build a part of Kepware Configuration API URL structure

    Returns the user_group specific url when a value is passed as the user_group name.
    '''
    
    if user_group == None:
        return USERGROUPS_ROOT
    else:
        return '{}/{}'.format(USERGROUPS_ROOT,user_group)

def add_user_group(server, DATA) -> Union[bool, list]:
    '''Add a "user_group" or multiple "user_group" objects to Kepware User Manager by passing a 
    list of user groups to be added all at once.

    INPUTS:
    "server" - instance of the "server" class

    "DATA" - properly JSON object (dict) of the user_group
    expected by Kepware Configuration API

    RETURNS:
    True - If a "HTTP 201 - Created" is received from Kepware

    List - If a "HTTP 207 - Multi-Status" is received from Kepware with a list of dict error responses for all 
    user groups added that failed.

    EXCEPTIONS:
    KepHTTPError - If urllib provides an HTTPError
    KepURLError - If urllib provides an URLError
    '''

    r = server._config_add(server.url + _create_url(), DATA)
    if r.code == 201: return True 
    elif r.code == 207:
        errors = [] 
        for item in r.payload:
            if item['code'] != 201:
                errors.append(item)
        return errors
    else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)

def del_user_group(server, user_group) -> bool:
    '''Delete a "user_group" object in Kepware User Manager
    
    INPUTS:
    "server" - instance of the "server" class

    "user_group" - name of user_group
    
    RETURNS:
    True - If a "HTTP 200 - OK" is received from Kepware

    EXCEPTIONS:
    KepHTTPError - If urllib provides an HTTPError
    KepURLError - If urllib provides an URLError
    '''

    r = server._config_del(server.url + _create_url(user_group))
    if r.code == 200: return True 
    else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)

def modify_user_group(server, DATA, user_group = None) -> bool:
    '''Modify a user_group object and it's properties in Kepware User Manager. If a "user_group" is not provided as an input,
    you need to identify the user_group in the 'common.ALLTYPES_NAME' property field in the "DATA". It will 
    assume that is the user_group that is to be modified.

    INPUTS:
    "server" - instance of the "server" class

    "DATA" - properly JSON object (dict) of the user_group properties to be modified.

    "user_group" (optional) - name of user_group to modify. Only needed if not existing in  "DATA"
    
    RETURNS:
    True - If a "HTTP 200 - OK" is received from Kepware

    EXCEPTIONS:
    KepHTTPError - If urllib provides an HTTPError
    KepURLError - If urllib provides an URLError
    '''
    
    # channel_data = server._force_update_check(force, DATA)
    if user_group == None:
        try:
            r = server._config_update(server.url + _create_url(DATA['common.ALLTYPES_NAME']), DATA)
            if r.code == 200: return True 
            else: return False
        except KeyError as err:
            err_msg = 'Error: No User Group identified in DATA | Key Error: {}'.format(err)
            raise KepError(err_msg)
        # except Exception as e:
        #     return 'Error: Error with {}: {}'.format(inspect.currentframe().f_code.co_name, str(e))
    else:
        r = server._config_update(server.url + _create_url(user_group), DATA)
        if r.code == 200: return True 
        else: return False

def get_user_group(server, user_group) -> dict:
    '''Returns the properties of the user_group object. Returned object is JSON.
    
    INPUTS:
    "server" - instance of the "server" class

    "user_group" - name of user_group
    
    RETURNS:
    dict - data for the user_group requested

    EXCEPTIONS:
    KepHTTPError - If urllib provides an HTTPError
    KepURLError - If urllib provides an URLError
    '''

    r = server._config_get(server.url + _create_url(user_group))
    return r.payload

def get_all_user_groups(server) -> list:
    '''Returns list of all user_group objects and their properties. Returned object is JSON list.
    
    INPUTS:
    "server" - instance of the "server" class
    
    RETURNS:
    list - data for all user_groups requested

    EXCEPTIONS:
    KepHTTPError - If urllib provides an HTTPError
    KepURLError - If urllib provides an URLError
    '''

    r = server._config_get(server.url + _create_url())
    return r.payload

def enable_user_group(server, user_group) -> bool:
    '''Enable the user group. Returned object is JSON.
    
    INPUTS:
    "server" - instance of the "server" class

    "user_group" - name of user group

    RETURNS:
    True - If a "HTTP 200 - OK" is received from Kepware

    EXCEPTIONS:
    KepHTTPError - If urllib provides an HTTPError
    KepURLError - If urllib provides an URLError
    '''
    DATA = {ENABLE_PROPERTY: True}
    return modify_user_group(server, DATA, user_group)

def disable_user_group(server, user_group) -> bool:
    '''Disable the user group. Returned object is JSON.
    
    INPUTS:
    "server" - instance of the "server" class

    "user_group" - name of user group

    RETURNS:
    True - If a "HTTP 200 - OK" is received from Kepware

    EXCEPTIONS:
    KepHTTPError - If urllib provides an HTTPError
    KepURLError - If urllib provides an URLError
    '''
    DATA = {ENABLE_PROPERTY: False}
    return modify_user_group(server, DATA, user_group)


# -------------------------------------------------------------------------
# Copyright (c), PTC Inc. and/or all its affiliates. All rights reserved.
# See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

r"""`users` exposes an API to allow modifications (add, delete, modify) to 
users within the Kepware Administration User Management through the Kepware Configuration API
"""
from typing import Union
from ..error import KepError, KepHTTPError


USERS_ROOT = '/admin/server_users'
ENABLE_PROPERTY = 'libadminsettings.USERMANAGER_USER_ENABLED'

def _create_url(user = None):
    '''Creates url object for the "server_users" branch of Kepware's project tree. Used 
    to build a part of Kepware Configuration API URL structure

    Returns the user specific url when a value is passed as the user name.
    '''
    
    if user == None:
        return USERS_ROOT
    else:
        return '{}/{}'.format(USERS_ROOT,user)

def add_user(server, DATA) -> Union[bool, list]:
    '''Add a "user" or multiple "user" objects to Kepware User Manager by passing a 
    list of users to be added all at once.

    INPUTS:
    "server" - instance of the "server" class

    "DATA" - properly JSON object (dict) of the user
    expected by Kepware Configuration API

    RETURNS:
    True - If a "HTTP 201 - Created" is received from Kepware

    List - If a "HTTP 207 - Multi-Status" is received from Kepware with a list of dict error responses for all 
    users added that failed.

    EXCEPTIONS:
    KepHTTPError - If urllib provides an HTTPError
    KepURLError - If urllib provides an URLError
    '''

    r = server._config_add(server.url + _create_url(), DATA)
    if r.code == 201: return True 
    elif r.code == 207:
        errors = [] 
        for item in r.payload:
            if item['code'] != 201:
                errors.append(item)
        return errors
    else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)

def del_user(server, user) -> bool:
    '''Delete a "user" object in Kepware User Manager
    
    INPUTS:
    "server" - instance of the "server" class

    "user" - name of user
    
    RETURNS:
    True - If a "HTTP 200 - OK" is received from Kepware

    EXCEPTIONS:
    KepHTTPError - If urllib provides an HTTPError
    KepURLError - If urllib provides an URLError
    '''

    r = server._config_del(server.url + _create_url(user))
    if r.code == 200: return True 
    else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)

def modify_user(server, DATA, user = None) -> bool:
    '''Modify a user object and it's properties in Kepware User Manager. If a "user" is not provided as an input,
    you need to identify the user in the 'common.ALLTYPES_NAME' property field in the "DATA". It will 
    assume that is the user that is to be modified.

    INPUTS:
    "server" - instance of the "server" class

    "DATA" - properly JSON object (dict) of the user properties to be modified.

    "user" (optional) - name of user to modify. Only needed if not existing in  "DATA"
    
    RETURNS:
    True - If a "HTTP 200 - OK" is received from Kepware

    EXCEPTIONS:
    KepHTTPError - If urllib provides an HTTPError
    KepURLError - If urllib provides an URLError
    '''
    
    # channel_data = server._force_update_check(force, DATA)
    if user == None:
        try:
            r = server._config_update(server.url + _create_url(DATA['common.ALLTYPES_NAME']), DATA)
            if r.code == 200: return True 
            else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)
        except KeyError as err:
            err_msg = 'Error: No User identified in DATA | Key Error: {}'.format(err)
            raise KepError(err_msg)
        # except Exception as e:
        #     return 'Error: Error with {}: {}'.format(inspect.currentframe().f_code.co_name, str(e))
    else:
        r = server._config_update(server.url + _create_url(user), DATA)
        if r.code == 200: return True 
        else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)

def get_user(server, user) -> dict:
    '''Returns the properties of the user object. Returned object is JSON.
    
    INPUTS:
    "server" - instance of the "server" class

    "user" - name of user
    
    RETURNS:
    dict - data for the user requested

    EXCEPTIONS:
    KepHTTPError - If urllib provides an HTTPError
    KepURLError - If urllib provides an URLError
    '''

    r = server._config_get(server.url + _create_url(user))
    return r.payload

def get_all_users(server) -> list:
    '''Returns list of all user objects and their properties. Returned object is JSON list.
    
    INPUTS:
    "server" - instance of the "server" class
    
    RETURNS:
    list - data for all users requested

    EXCEPTIONS:
    KepHTTPError - If urllib provides an HTTPError
    KepURLError - If urllib provides an URLError
    '''

    r = server._config_get(server.url + _create_url())
    return r.payload

def enable_user(server, user) -> bool:
    '''Enable the user. Returned object is JSON.
    
    INPUTS:
    "server" - instance of the "server" class

    "user" - name of user

    RETURNS:
    True - If a "HTTP 200 - OK" is received from Kepware

    EXCEPTIONS:
    KepHTTPError - If urllib provides an HTTPError
    KepURLError - If urllib provides an URLError
    '''
    DATA = {ENABLE_PROPERTY: True}
    return modify_user(server, DATA, user)

def disable_user(server, user) -> bool:
    '''Disable the user. Returned object is JSON.
    
    INPUTS:
    "server" - instance of the "server" class

    "user" - name of user

    RETURNS:
    True - If a "HTTP 200 - OK" is received from Kepware

    EXCEPTIONS:
    KepHTTPError - If urllib provides an HTTPError
    KepURLError - If urllib provides an URLError
    '''
    DATA = {ENABLE_PROPERTY: False}
    return modify_user(server, DATA, user)


# -------------------------------------------------------------------------
# Copyright (c) PTC Inc. and/or all its affiliates. All rights reserved.
# See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------


r"""`exchange` exposes an API to allow modifications (add, delete, modify) to 
exchange objects for EGD devices within the Kepware Configuration API
"""

import kepconfig
from typing import Union
from .. import egd as EGD, channel, device

CONSUMER_ROOT = '/consumer_exchange_groups/consumer exchanges/consumer_exchanges'
PRODUCER_ROOT = '/producer_exchange_groups/producer exchanges/producer_exchanges'

def _create_url(device_path, ex_type, exchange_name = None):
    '''Creates url object for the "exchange" branch of Kepware's project tree. Used 
    to build a part of Kepware Configuration API URL structure

    Returns the exchange specific url when a value is passed as the exchange name.
    '''
    path_obj = kepconfig.path_split(device_path)
    device_root = channel._create_url(path_obj['channel']) + device._create_url(path_obj['device'])

    if exchange_name == None:
        if ex_type == EGD.CONSUMER_EXCHANGE:
            return device_root + CONSUMER_ROOT
        else:
            return device_root + PRODUCER_ROOT
    else:
        if ex_type == EGD.CONSUMER_EXCHANGE:
            return '{}{}/{}'.format(device_root,CONSUMER_ROOT,exchange_name)
        else:
            return '{}{}/{}'.format(device_root,PRODUCER_ROOT,exchange_name)

def add_exchange(server, device_path, ex_type, DATA) -> Union[bool, list]:
    '''Add a "exchange" or multiple "exchange" objects to Kepware. Can be used to pass children of a exchange object 
    such as ranges. This allows you to create a exchange and ranges for the exchange all in one function, if desired.

    Additionally it can be used to pass a list of exchanges and it's children to be added all at once.

    INPUTS:

    "server" - instance of the "server" class

    "device_path" - path to exchanges. Standard Kepware address decimal 
    notation string such as "channel1.device1"

    "type" - type of exchange either consumer or producer

    "DATA" - properly JSON object (dict) of the exchange and it's children
    expected by Kepware Configuration API

    RETURNS:

    True - If a "HTTP 201 - Created" is received from Kepware
    
    List - If a "HTTP 207 - Multi-Status" is received from Kepware with a list of dict error responses for all 
    exchanges added that failed.

    False - If a non-expected "2xx successful" code is returned
        

    EXCEPTIONS:

    KepHTTPError - If urllib provides an HTTPError
    KepURLError - If urllib provides an URLError
    '''

    r = server._config_add(server.url + _create_url(device_path, ex_type), DATA)
    if r.code == 201: return True
    elif r.code == 207:
        errors = [] 
        for item in r.payload:
            if item['code'] != 201:
                errors.append(item)
        return errors
    else: return False

def del_exchange(server, device_path, ex_type, exchange_name) -> bool:
    '''Delete a "exchange" object in Kepware. This will delete all children as well
    
    INPUTS:

    "server" - instance of the "server" class

    "device_path" - path to exchanges. Standard Kepware address decimal 
    notation string such as "channel1.device1"

    "ex_type" - type of exchange either consumer or producer

    "exchange_name" - name of exchange
    
    RETURNS:

    True - If a "HTTP 200 - OK" is received from Kepware

    EXCEPTIONS:

    KepHTTPError - If urllib provides an HTTPError
    KepURLError - If urllib provides an URLError
    '''

    r = server._config_del(server.url + _create_url(device_path, ex_type, exchange_name))
    if r.code == 200: return True 
    else: return False

def modify_exchange(server, device_path, ex_type, DATA, exchange_name = None, force = False) -> bool:
    '''Modify a exchange object and it's properties in Kepware. If a "exchange_name" is not provided as an input,
    you need to identify the exchange in the 'common.ALLTYPES_NAME' property field in the "DATA". It will 
    assume that is the exchange that is to be modified.

    INPUTS:

    "server" - instance of the "server" class

    "device_path" - path to exchanges. Standard Kepware address decimal 
    notation string such as "channel1.device1"

    "ex_type" - type of exchange either consumer or producer

    "DATA" - properly JSON object (dict) of the exchange properties to be modified.

    "exchange_name" (optional) - name of exchange to modify. Only needed if not existing in  "DATA"

    "force" (optional) - if True, will force the configuration update to the Kepware server
    
    RETURNS:
    
    True - If a "HTTP 200 - OK" is received from Kepware

    EXCEPTIONS:
    
    KepHTTPError - If urllib provides an HTTPError
    KepURLError - If urllib provides an URLError
    '''
    
    exchange_data = server._force_update_check(force, DATA)
    if exchange_name == None:
        try:
            r = server._config_update(server.url + _create_url(device_path, ex_type, exchange_data['common.ALLTYPES_NAME']), exchange_data)
            if r.code == 200: return True 
            else: return False
        except KeyError as err:
            print('Error: No exchange identified in DATA | Key Error: {}'.format(err))
            return False
        # except Exception as e:
        #     return 'Error: Error with {}: {}'.format(inspect.currentframe().f_code.co_name, str(e))
    else:
        r = server._config_update(server.url + _create_url(device_path, ex_type, exchange_name), exchange_data)
        if r.code == 200: return True 
        else: return False

def get_exchange(server, device_path, ex_type, exchange_name = None) -> Union[dict, list]:
    '''Returns the properties of the exchange object or a list of all exchanges and their 
    properties for the type input. Returned object is JSON.
    
    INPUTS:

    "server" - instance of the "server" class

    "device_path" - path to exchanges. Standard Kepware address decimal 
    notation string such as "channel1.device1"

    "ex_type" - type of exchange either consumer or producer

    "exchange_name" - name of exchange
    
    RETURNS:

    JSON - data for the exchange requested or a list of exchanges and their properties

    EXCEPTIONS:

    KepHTTPError - If urllib provides an HTTPError
    KepURLError - If urllib provides an URLError
    '''
    if exchange_name == None:
        r = server._config_get(server.url + _create_url(device_path, ex_type))
    else:
        r = server._config_get(server.url + _create_url(device_path, ex_type, exchange_name))
    return r.payload

def get_all_exchanges(server, device_path):
    '''Returns list of all exchange objects and their properties. Returned object is JSON list.
    
    INPUTS:

    "server" - instance of the "server" class

    "device_path" - path to exchanges. Standard Kepware address decimal 
    notation string such as "channel1.device1"
    
    RETURNS:

    List - [list of consumer exchanges, list of producer exchanges] - list of lists for all 
    exchanges for the device

    EXCEPTIONS:

    KepHTTPError - If urllib provides an HTTPError
    KepURLError - If urllib provides an URLError
    '''
    exchange_list = []
    exchange_list.append(get_exchange(server, device_path, EGD.CONSUMER_EXCHANGE))
    exchange_list.append(get_exchange(server, device_path, EGD.PRODUCER_EXCHANGE))
    return exchange_list


# -------------------------------------------------------------------------
# Copyright (c) PTC Inc. and/or all its affiliates. All rights reserved.
# See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

r"""`connection` exposes an API that manages the RESTful requests 
for the Kepware Configuration API. 
"""

import json
import codecs
import datetime
from urllib import request, parse, error
from base64 import b64encode
from .error import KepError, KepHTTPError, KepURLError
import socket
import ssl
import sys

class KepServiceResponse:
    '''A class to represent a return object when calling a "service" API of Kepware. This is
    used to return the responses when a "service" is executed appropriately

    Properties:

    "code" - HTTP code returned
    "message" - return from the "service" call
    "href" - URL reference to the JOB that is created by the service API
    '''

    def __init__(self, code = '', message = '', href = ''):
        self.code = code
        self.message = message
        self.href = href
    
    def __str__(self):
        return '{"code": %s, "message": %s, "href": %s}' % (self.code, self.message, self.href)

class _HttpDataAbstract:
    def __init__(self):
        self.payload = ''
        self.code = ''
        self.reason = ''

class server:
    '''A class to represent a connection to an instance of Kepware. This object is used to 
    create the Authentication and server parameters to taget a Kepware instance. An instance of this is 
    used in all configuration calls done.

    Properties:

    "host" - host name or IP address
    "port" - port of Configuration API
    "username" - username to conduct "Basic Authentication"
    "password" - password to conduct "Basic Authentication"
    "url" (STATIC) - base URL for the server connection
    "SSL_on" - Identify to use HTTPS connection (Default: False)
    "SSL_ignore_hostname" - During certificate validation ignore the hostname check
    "SSL_trust_all_certs" (insecure) - During certificate validation trust any certificate - if True, 
        will "set SSL_ignore_hostname" to true

    Methods:

    "reinitialize()" - Reinitialize the Kepware server
    "get_trans_log()" - retrieve the Configuration API transaction logs
    "get_event_log()" - retrieve the Kepware Event Log
    '''
    __root_url = '/config'
    __version_url = '/v1'
    __project_services_url = '/project/services'
    __event_log_url = '/event_log'
    __trans_log_url = '/log'



    def __init__(self,  host = None, port = None, user = None, pw = None, https = False):
        self.host = host
        self.port = port
        self.username = user
        self.password = pw
        self.__ssl_context = ssl.create_default_context()
        self.__SSL_on = https
    
    @property
    def url(self):
        if self.SSL_on:
            proto = 'https'
        else:
            proto = 'http'
        return  '{}://{}:{}{}{}'.format(proto, self.host, self.port, self.__root_url, self.__version_url)
    
    @property
    def SSL_on(self):
        return self.__SSL_on
    
    @SSL_on.setter
    def SSL_on(self, val):
        if isinstance(val, bool):
            self.__SSL_on = val

    @property
    def SSL_ignore_hostname(self):
        return not self.__ssl_context.check_hostname

    @SSL_ignore_hostname.setter
    def SSL_ignore_hostname(self, val):
        if isinstance(val, bool):
            if val == True:
                self.__ssl_context.check_hostname = False
            else:
                self.__ssl_context.check_hostname = True

    
    @property
    def SSL_trust_all_certs(self):
        if self.__ssl_context.verify_mode == ssl.CERT_NONE:
            return True
        else:
            return False

    @SSL_trust_all_certs.setter
    def SSL_trust_all_certs(self, val):
        if isinstance(val, bool):
            if val == True:
                if self.__ssl_context.check_hostname == True:
                    self.__ssl_context.check_hostname = False
                self.__ssl_context.verify_mode = ssl.CERT_NONE
            else:
                self.__ssl_context.verify_mode = ssl.CERT_REQUIRED




    def reinitialize(self, job_ttl = None) -> KepServiceResponse:
        '''Executes a Reinitialize call to the Kepware instance.

        RETURNS:
        KepServiceResponse instance with job information

        EXCEPTIONS (If not HTTP 200 or 429 returned):
        
        KepHTTPError - If urllib provides an HTTPError
        KepURLError - If urllib provides an URLError
        '''
        url = self.url + self.__project_services_url + '/ReinitializeRuntime' 
        try:
            job = self._kep_service_execute(url,job_ttl)
            return job
        except Exception as err:
            raise err
        
    def get_trans_log(self, start = None, end = None, limit = None) -> list:
        ''' Get the Transaction Log from the Kepware instance.

        "start" (optional) - datetime.datetime type and should be UTC

        "end" (optional) - datetime.datetime type and should be UTC

        "limit" (optional) -  number of event log entries to request
        '''
        query = self.__create_query(start, end, limit)
        url = self.url + self.__trans_log_url + '?' + parse.urlencode(query)
        r = self._config_get(url)
        return r.payload

    def get_event_log(self, limit = None, start = None, end = None) -> list:
        ''' Get the Event Log from the Kepware instance.

        "start" (optional) - datetime.datetime type and should be UTC

        "end" (optional) - datetime.datetime type and should be UTC

        "limit" (optional) - number of event log entries to request
        '''
        query = self.__create_query(start, end, limit)
        url = self.url + self.__event_log_url + '?' + parse.urlencode(query)
        r = self._config_get(url)
        return r.payload
    
    def get_project_properties(self) -> dict:
        ''' Get the Project Properties of the Kepware instance.
        
        RETURNS:
        dict - Dict of all the project properties

        EXCEPTIONS:
        KepHTTPError - If urllib provides an HTTPError
        KepURLError - If urllib provides an URLError
        '''

        r = self._config_get(self.url + '/project')
        return r.payload
    
    def modify_project_properties(self, DATA, force = False) -> bool:
        ''' Modify the Project Properties of the Kepware instance.

        INPUTS:

        "DATA" - properly JSON object (dict) of the project properties to be modified

        "force" (optional) - if True, will force the configuration update to the Kepware server
        
        RETURNS:
        True - If a "HTTP 200 - OK" is received from Kepware

        EXCEPTIONS:
        KepHTTPError - If urllib provides an HTTPError
        KepURLError - If urllib provides an URLError
        '''

        prop_data = self._force_update_check(force, DATA)
        r = self._config_update(self.url + '/project', prop_data)
        if r.code == 200: return True 
        else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)

    #Function used to Add an object to Kepware (HTTP POST)
    def _config_add(self, url, DATA):
        '''Conducts an POST method at *url* to add an object in the Kepware Configuration
        *DATA* is required to be a properly JSON object (dict) of the item to be posted to *url* 
        '''
        data = json.dumps(DATA).encode('utf-8')
        url_obj = self.__url_validate(url)
        q = request.Request(url_obj, data, method='POST')
        r = self.__connect(q)
        return r

    #Function used to del an object to Kepware (HTTP DELETE)
    def _config_del(self, url):
        '''Conducts an DELETE method at *url* to delete an object in the Kepware Configuration'''
        url_obj = self.__url_validate(url)
        q = request.Request(url_obj, method='DELETE')
        r = self.__connect(q)
        return r

    #Function used to Update an object to Kepware (HTTP PUT)
    def _config_update(self, url, DATA = None):
        '''Conducts an PUT method at *url* to modify an object in the Kepware Configuration.
        *DATA* is required to be a properly JSON object (dict) of the item to be put to *url*
        '''
        url_obj = self.__url_validate(url)
        if DATA == None:            
            q = request.Request(url_obj, method='PUT')
        else:
            data = json.dumps(DATA).encode('utf-8')
            q = request.Request(url_obj, data, method='PUT')
        r = self.__connect(q)
        return r

    #Function used to Read an object from Kepware (HTTP GET) and return the JSON response
    def _config_get(self, url):
        '''Conducts an GET method at *url* to retrieve an objects properties in the Kepware Configuration.'''
        url_obj = self.__url_validate(url)
        q = request.Request(url_obj, method='GET')
        r = self.__connect(q)
        return r

    
    def _force_update_check(self, force, DATA):
        if force == True:
            DATA['FORCE_UPDATE'] = True
        else:
            # Get Project ID if it doesn't exist and if FORCE_UPDATE is existing and FALSE
            if 'PROJECT_ID' not in DATA:
                if 'FORCE_UPDATE' in DATA:
                    if 'FORCE_UPDATE' == False:
                        try:
                            project_data = self._config_get(self.url + '/project')
                            DATA['PROJECT_ID'] = project_data.payload['PROJECT_ID']
                        except:
                            #NEED TO COVER ERROR CONDITION
                            pass
                else:
                    try:
                        project_data = self._config_get(self.url + '/project')
                        DATA['PROJECT_ID'] = project_data.payload['PROJECT_ID']
                    except:
                        #NEED TO COVER ERROR CONDITION
                        pass
        return DATA
    # General service call handler
    def _kep_service_execute(self, url, TTL):
        try:
            if TTL != None:
                TTL = {"servermain.JOB_TIME_TO_LIVE_SECONDS": TTL}
            r = self._config_update(url, TTL)
            job = KepServiceResponse(r.payload['code'],r.payload['message'], r.payload['href'])
            return job
        except KepHTTPError as err:
            if err.code == 429:
                job = KepServiceResponse()
                job.code = err.code
                job.message = err.payload
                return job
            else:
                raise err

# 
# Supporting Functions
#

    # General connect call to manage HTTP responses for all methods
    # Returns the response object for the method to handle as appropriate
    # Raises Errors as found
    def __connect(self,request_obj):
        # Fill appropriate header information
        data = _HttpDataAbstract()
        request_obj.add_header("Authorization", "Basic %s" % self.__build_auth_str(self.username, self.password))
        request_obj.add_header("Content-Type", "application/json")
        request_obj.add_header("Accept", "application/json")
        try:
            # context is sent regardless of HTTP or HTTPS - seems to be ignored if HTTP URL
            with request.urlopen(request_obj, context=self.__ssl_context) as server:
                try:
                    payload = server.read()
                    data.payload = json.loads(codecs.decode(payload,'utf-8-sig'))
                except:
                    pass
                data.code = server.code
                data.reason = server.reason
                return data
        except error.HTTPError as err:
            payload = json.loads(codecs.decode(err.read(),'utf-8-sig'))
            # print('HTTP Code: {}\n{}'.format(err.code,payload), file=sys.stderr)
            raise KepHTTPError(err.url, err.code, err.msg, err.hdrs, payload)
        except error.URLError as err:
            # print('URLError: {} URL: {}'.format(err.reason, request_obj.get_full_url()), file=sys.stderr)
            raise KepURLError(err.reason, request_obj.get_full_url())

    # Fucntion used to ensure special characters are handled in the URL
    # Ex: Space will be turned to %20
    def __url_validate(self, url):
        parsed_url = parse.urlparse(url)
        # Added % for scenarios where special characters have already been escaped with %
        updated_path = parse.quote(parsed_url.path, safe = '/%')

        # If host is "localhost", force using the IPv4 loopback adapter IP address in all calls
        # This is done to remove retries that will happen when the host resolution uses IPv6 intially
        # Kepware currently doesn't support IPv6 and is not listening on this interface
        if parsed_url.hostname.lower() == 'localhost':
            ip = socket.gethostbyname(parsed_url.hostname)
            parsed_url = parsed_url._replace(netloc='{}:{}'.format(ip, parsed_url.port))
        
        return parsed_url._replace(path=updated_path).geturl()

    # Function used to build the basic authentication string
    def __build_auth_str(self, username, password):
        if isinstance(username, str):
            username = username.encode('latin1')
        if isinstance(password, str):
            password = password.encode('latin1')
        authstr = b64encode(b':'.join((username, password))).strip().decode('ascii')
        return authstr
    
    # Create parameters for log queries
    def __create_query(self, start = None, end = None, limit = None):
        query = {}
        if start != None and isinstance(start, datetime.datetime):
            query['start'] = start.isoformat()
        if end != None and isinstance(end, datetime.datetime):
            query['end'] = end.isoformat()
        if limit != None:
            query['limit'] = limit
        return query





# -------------------------------------------------------------------------
# Copyright (c) PTC Inc. and/or all its affiliates. All rights reserved.
# See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------


r"""`names` exposes an API to allow modifications (add, delete, modify) to 
name resolution objects for EGD devices within the Kepware Configuration API
"""

import kepconfig
from typing import Union
from .. import channel, device

NAMES_ROOT = '/name_resolution_groups/Name Resolutions/name_resolutions'

def _create_url(device_path, name = None):
    '''Creates url object for the "name resolution" branch of Kepware's project 
    tree. Used to build a part of Kepware Configuration API URL structure

    Returns the name resolution specific url when a value is passed.
    '''
    path_obj = kepconfig.path_split(device_path)
    device_root = channel._create_url(path_obj['channel']) + device._create_url(path_obj['device'])

    if name == None:
        return '{}/{}'.format(device_root, NAMES_ROOT)
    else:
        return '{}/{}/{}'.format(device_root, NAMES_ROOT, name)

def add_name_resolution(server, device_path, DATA) -> Union[bool, list]:
    '''Add a "name resolution" or multiple "name resolution" objects to Kepware. This allows you to 
    create a name resolution or multiple name resolutions all in one function, if desired.

    INPUTS:

    "server" - instance of the "server" class

    "device_path" - path to name resolutions. Standard Kepware address decimal 
    notation string such as "channel1.device1"

    "DATA" - properly JSON object (dict) of the range or ranges

    RETURNS:

    True - If a "HTTP 201 - Created" is received from Kepware
    
    List - If a "HTTP 207 - Multi-Status" is received from Kepware with a list of dict error responses for all 
    name resolutions added that failed.

    False - If a non-expected "2xx successful" code is returned

    EXCEPTIONS:

    KepHTTPError - If urllib provides an HTTPError
    KepURLError - If urllib provides an URLError
    '''

    r = server._config_add(server.url + _create_url(device_path), DATA)
    if r.code == 201: return True
    elif r.code == 207:
        errors = [] 
        for item in r.payload:
            if item['code'] != 201:
                errors.append(item)
        return errors
    else: return False

def del_name_resolution(server, device_path, name) -> bool:
    '''Delete a "name resolution" object in Kepware.
    
    INPUTS:

    "server" - instance of the "server" class

    "device_path" - path to exchanges and their ranges. Standard Kepware address decimal 
    notation string such as "channel1.device1"

    "name resolution" - name of name resolution
    
    RETURNS:

    True - If a "HTTP 200 - OK" is received from Kepware

    EXCEPTIONS:

    KepHTTPError - If urllib provides an HTTPError
    KepURLError - If urllib provides an URLError
    '''

    r = server._config_del(server.url + _create_url(device_path, name))
    if r.code == 200: return True 
    else: return False

def modify_name_resolution(server, device_path, DATA, name = None, force = False) -> bool:
    '''Modify a name resolution object and it's properties in Kepware. If a "name" is not provided as an input,
    you need to identify the name resolution in the 'common.ALLTYPES_NAME' property field in the "DATA". It will 
    assume that is the name resolution that is to be modified.

    INPUTS:

    "server" - instance of the "server" class

    "device_path" - path to exchanges and their ranges. Standard Kepware address decimal 
    notation string such as "channel1.device1"

    "DATA" - properly JSON object (dict) of the range properties to be modified.

    "name" (optional) - name of name resolution to modify. Only needed if not existing in  "DATA"

    "force" (optional) - if True, will force the configuration update to the Kepware server
    
    RETURNS:
    
    True - If a "HTTP 200 - OK" is received from Kepware

    EXCEPTIONS:
    
    KepHTTPError - If urllib provides an HTTPError
    KepURLError - If urllib provides an URLError
    '''
    
    name_data = server._force_update_check(force, DATA)
    if name == None:
        try:
            r = server._config_update(server.url + _create_url(device_path, name_data['common.ALLTYPES_NAME']), name_data)
            if r.code == 200: return True 
            else: return False
        except KeyError as err:
            print('Error: No name resolution identified in DATA | Key Error: {}'.format(err))
            return False
        # except Exception as e:
        #     return 'Error: Error with {}: {}'.format(inspect.currentframe().f_code.co_name, str(e))
    else:
        r = server._config_update(server.url + _create_url(device_path, name), name_data)
        if r.code == 200: return True 
        else: return False

def get_name_resolution(server, device_path, name = None) -> Union[dict, list]:
    '''Returns the properties of the name resolution object or a list of all name resolutions.
    Returned object is JSON.
    
    INPUTS:

    "server" - instance of the "server" class

    "device_path" - path to exchanges and their ranges. Standard Kepware address decimal 
    notation string such as "channel1.device1"

    "name" - name of name resolution
    
    RETURNS:

    JSON - data for the name resolution requested or a list of name resolutions and their properties

    EXCEPTIONS:

    KepHTTPError - If urllib provides an HTTPError
    KepURLError - If urllib provides an URLError
    '''

    if name == None:
        r = server._config_get(server.url + _create_url(device_path))
    else:
        r = server._config_get(server.url + _create_url(device_path, name))
    return r.payload


# -------------------------------------------------------------------------
# Copyright (c) PTC Inc. and/or all its affiliates. All rights reserved.
# See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------


r"""`tag` exposes an API to allow modifications (add, delete, modify) to 
tag and tag group objects within the Kepware Configuration API
"""

from ..error import KepError, KepHTTPError
from typing import Union
import kepconfig
from . import channel, device
import inspect

TAGS_ROOT = '/tags'
TAG_GRP_ROOT = '/tag_groups'

def _create_tags_url(tag = None):
    '''Creates url object for the "tags" branch of Kepware's project tree. Used 
    to build a part of Kepware Configuration API URL structure
    
    Returns the tag specific url when a value is passed as the tag name.
    '''
    if tag == None:
        return TAGS_ROOT
    else: 
        return '{}/{}'.format(TAGS_ROOT,tag)

def _create_tag_groups_url(tag_group = None):
    '''Creates url object for the "tag_group" branch of Kepware's project tree. Used 
    to build a part of Kepware Configuration API URL structure
    
    Returns the tag group specific url when a value is passed as the tag group name.
    '''
    if tag_group == None:
        return TAG_GRP_ROOT
    else: 
        return '{}/{}'.format(TAG_GRP_ROOT,tag_group)

def add_tag(server, tag_path, DATA) -> Union[bool, list]:
    '''Add "tag" objects to a specific path in Kepware. To be used to 
    pass a list of tags to be added at one path location.

    INPUTS:
    "server" - instance of the "server" class
    
    "tag_path" -  path identifying location to add tags. Standard Kepware address decimal 
    notation string such as "channel1.device1.tag_group1"
    
    "DATA" - properly JSON object (dict) of the tags 
    expected by Kepware Configuration API at the "tags" url

    RETURNS:
    True - If a "HTTP 201 - Created" is received from Kepware

    List  - If a "HTTP 207 - Multi-Status" is received from Kepware with a list of dict error responses for all 
    tags added that failed.

    EXCEPTIONS:
    KepHTTPError - If urllib provides an HTTPError
    KepURLError - If urllib provides an URLError
    '''

    path_obj = kepconfig.path_split(tag_path)
    try:
        url = server.url+channel._create_url(path_obj['channel'])+device._create_url(path_obj['device'])
        if 'tag_path' in path_obj:
            for tg in path_obj['tag_path']:
                url += _create_tag_groups_url(tag_group=tg)
        url += _create_tags_url()
    except KeyError as err:
        err_msg = 'Error: No key {} identified | Function: {}'.format(err, inspect.currentframe().f_code.co_name)
        raise KepError(err_msg)
    except Exception as e:
        err_msg = 'Error: Error with {}: {}'.format(inspect.currentframe().f_code.co_name, str(e))
        raise KepError(err_msg)
    r = server._config_add(url, DATA)
    if r.code == 201: return True 
    elif r.code == 207:
        errors = [] 
        for item in r.payload:
            if item['code'] != 201:
                errors.append(item)
        return errors
    else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)

def add_tag_group(server, tag_group_path, DATA) -> Union[bool, list]:
    '''Add "tag_group" objects to a specific path in Kepware. To be used to 
    pass a list of tag_groups and children (tags or tag groups) to be added at one 
    path location.

    INPUTS:
    "server" - instance of the "server" class
    
    "tag_group_path" - path identifying location to add tag groups. Standard Kepware address decimal 
    notation string such as "channel1.device1.tag_group1"
    
    "DATA" - properly JSON object (dict) of the tag groups
    and it's children expected by Kepware Configuration API at the "tag_groups" url

    RETURNS:
    True - If a "HTTP 201 - Created" is received from Kepware

    List  - If a "HTTP 207 - Multi-Status" is received from Kepware with a list of dict error responses for all 
    tag groups added that failed.

    EXCEPTIONS:
    KepHTTPError - If urllib provides an HTTPError
    KepURLError - If urllib provides an URLError
    '''

    path_obj = kepconfig.path_split(tag_group_path)
    try:
        url = server.url+channel._create_url(path_obj['channel'])+device._create_url(path_obj['device'])
        if 'tag_path' in path_obj:
            for tg in path_obj['tag_path']:
                url += _create_tag_groups_url(tag_group=tg)
        url += _create_tag_groups_url()
    except KeyError as err:
        err_msg = 'Error: No key {} identified | Function: {}'.format(err, inspect.currentframe().f_code.co_name)
        raise KepError(err_msg)
    except Exception as e:
        err_msg = 'Error: Error with {}: {}'.format(inspect.currentframe().f_code.co_name, str(e))
        raise KepError(err_msg)
    r = server._config_add(url, DATA)
    if r.code == 201: return True 
    elif r.code == 207:
        errors = [] 
        for item in r.payload:
            if item['code'] != 201:
                errors.append(item)
        return errors
    else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)

def add_all_tags(server, ch_dev_path, DATA) -> Union[bool, list]:
    '''Add "tag" and "tag group" objects to a device in Kepware. To be used to 
    pass a list of tags, tag groups and/or children of tag groups (tags and tag 
    groups) to be added at once.
    
    INPUTS:
    "server" - instance of the "server" class

    "ch_dev_path" - path to add tags and tag groups. Standard Kepware address decimal 
    notation string such as "channel1.device1"
    
    "DATA" - properly JSON object (dict) of the tags, 
    tag groups and it's children expected by Kepware Configuration API.
    
    RETURNS:
    True - If a "HTTP 201 - Created" is received from Kepware for all items

    List  - [tag failure list, tag group failure list] -   If a "HTTP 207 - Multi-Status" is received from 
    Kepware for either tags or tag groups, a list of dict error responses for all tags and/or tag groups added that failed. 

    False - If tags or tag groups are not found in DATA

    EXCEPTIONS:
    KepHTTPError - If urllib provides an HTTPError
    KepURLError - If urllib provides an URLError
    '''
######################################################
# Need to Handle HTTP 207 from the tag/tag group calls
######################################################
    
    tags_result = False
    tag_groups_result = False

    # check to see if there are dict entries for tags or tag groups
    if ('tag_groups' not in DATA) and ('tags' not in DATA):
        return False

    if 'tags' in DATA:
        tags_result = add_tag(server, ch_dev_path, DATA['tags'])
    if 'tag_groups' in DATA:
        #Add all Tag Groups
        tag_groups_result = add_tag_group(server, ch_dev_path, DATA['tag_groups'])
    
    # build results return from both calls
    if tags_result == True and tag_groups_result == True:
        return True
    elif tags_result == True:
        return [[], tag_groups_result]
    elif tag_groups_result == True:
        return [tags_result, []]
    else:
        # mixed results from both tags and tag groups
        return [tags_result, tag_groups_result]

def modify_tag(server, full_tag_path, DATA, force = False) -> bool:
    '''Modify a "tag" object and it's properties in Kepware.

    INPUTS:
    "server" - instance of the "server" class
    
    "full_tag_path" - path identifying location and tag to modify. Standard Kepware address decimal 
    notation string including the tag such as "channel1.device1.tag_group1.tag1"

    "DATA" - properly JSON object (dict) of the tag properties to be modified.

    "force" (optional) - if True, will force the configuration update to the Kepware server

    RETURNS:
    True - If a "HTTP 200 - OK" is received from Kepware

    EXCEPTIONS:
    KepHTTPError - If urllib provides an HTTPError
    KepURLError - If urllib provides an URLError
    '''

    tag_data = server._force_update_check(force, DATA)

    path_obj = kepconfig.path_split(full_tag_path)
    try:
        url = server.url+channel._create_url(path_obj['channel'])+device._create_url(path_obj['device'])
        for x in range(0, len(path_obj['tag_path'])-1):
            url += _create_tag_groups_url(tag_group=path_obj['tag_path'][x])
        url += _create_tags_url(tag=path_obj['tag_path'][len(path_obj['tag_path'])-1])
    except KeyError as err:
        err_msg = 'Error: No key {} identified | Function: {}'.format(err, inspect.currentframe().f_code.co_name)
        raise KepError(err_msg)
    except Exception as e:
        err_msg = 'Error: Error with {}: {}'.format(inspect.currentframe().f_code.co_name, str(e))
        raise KepError(err_msg)
    r = server._config_update(url, tag_data)
    if r.code == 200: return True 
    else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)

def modify_tag_group(server, tag_group_path, DATA, force = False) -> bool:
    '''Modify a "tag group" object and it's properties in Kepware.

    INPUTS:
    "server" - instance of the "server" class
    
    "tag_group_path" - path identifying location and tag group to modify. Standard Kepware address decimal 
    notation string including the tag such as "channel1.device1.tag_group1"

    "DATA" is required to be a properly JSON object (dict) of the tag properties to be modified.

    RETURNS:
    True - If a "HTTP 200 - OK" is received from Kepware

    EXCEPTIONS:
    KepHTTPError - If urllib provides an HTTPError
    KepURLError - If urllib provides an URLError
    '''

    tag_group_data = server._force_update_check(force, DATA)

    path_obj = kepconfig.path_split(tag_group_path)
    try:
        url = server.url+channel._create_url(path_obj['channel'])+device._create_url(path_obj['device'])
        for tg in path_obj['tag_path']:
            url += _create_tag_groups_url(tag_group=tg)
    except KeyError as err:
        err_msg = 'Error: No key {} identified | Function: {}'.format(err, inspect.currentframe().f_code.co_name)
        raise KepError(err_msg)
    except Exception as e:
        err_msg = 'Error: Error with {}: {}'.format(inspect.currentframe().f_code.co_name, str(e))
        raise KepError(err_msg)
    r = server._config_update(url, tag_group_data)
    if r.code == 200: return True 
    else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)

def del_tag(server, full_tag_path) -> bool:
    '''Delete "tag" object at a specific path in Kepware.

    INPUTS:
    "server" - instance of the "server" class
    
    "full_tag_path" - path identifying location and tag to delete. Standard Kepware address decimal 
    notation string including the tag such as "channel1.device1.tag_group1.tag1"

    RETURNS:
    True - If a "HTTP 200 - OK" is received from Kepware

    EXCEPTIONS:
    KepHTTPError - If urllib provides an HTTPError
    KepURLError - If urllib provides an URLError
    '''

    path_obj = kepconfig.path_split(full_tag_path)
    try:
        url = server.url+channel._create_url(path_obj['channel'])+device._create_url(path_obj['device'])
        for x in range(0, len(path_obj['tag_path'])-1):
            url += _create_tag_groups_url(tag_group=path_obj['tag_path'][x])
        url += _create_tags_url(tag=path_obj['tag_path'][len(path_obj['tag_path'])-1])
    except KeyError as err:
        err_msg = 'Error: No key {} identified | Function: {}'.format(err, inspect.currentframe().f_code.co_name)
        raise KepError(err_msg)
    except Exception as e:
        err_msg = 'Error: Error with {}: {}'.format(inspect.currentframe().f_code.co_name, str(e))
        raise KepError(err_msg)
    r = server._config_del(url)
    if r.code == 200: return True 
    else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)

def del_tag_group(server, tag_group_path) -> bool:
    '''Delete "tag group" object at a specific path in Kepware.

    INPUTS:
    "server" - instance of the "server" class
    
    "tag_group_path" - path identifying location and tag group to delete. Standard Kepware address decimal 
    notation string such as "channel1.device1.tag_group1"

    RETURNS:
    True - If a "HTTP 200 - OK" is received from Kepware

    EXCEPTIONS:
    KepHTTPError - If urllib provides an HTTPError
    KepURLError - If urllib provides an URLError
    '''

    path_obj = kepconfig.path_split(tag_group_path)
    try:
        url = server.url+channel._create_url(path_obj['channel'])+device._create_url(path_obj['device'])
        for tg in path_obj['tag_path']:
            url += _create_tag_groups_url(tag_group=tg)
    except KeyError as err:
        err_msg = 'Error: No key {} identified | Function: {}'.format(err, inspect.currentframe().f_code.co_name)
        raise KepError(err_msg)
    except Exception as err:
        err_msg = 'Error: Error with {}: {}'.format(inspect.currentframe().f_code.co_name, str(err))
        raise KepError(err_msg)
    r = server._config_del(url)
    if r.code == 200: return True 
    else: raise KepHTTPError(r.url, r.code, r.msg, r.hdrs, r.payload)

def get_tag(server, full_tag_path) -> dict:
    '''Returns the properties of the "tag" object at a specific path in Kepware. 
    Returned object is JSON. 

    INPUTS:
    "server" - instance of the "server" class
    
    "full_tag_path" - path identifying tag. Standard Kepware address decimal 
    notation string including the tag such as "channel1.device1.tag_group1.tag1"

    RETURNS:
    dict - data for the tag requested

    EXCEPTIONS:
    KepHTTPError - If urllib provides an HTTPError
    KepURLError - If urllib provides an URLError
    '''

    path_obj = kepconfig.path_split(full_tag_path)
    try:
        url = server.url+channel._create_url(path_obj['channel'])+device._create_url(path_obj['device'])
        for x in range(0, len(path_obj['tag_path'])-1):
            url += _create_tag_groups_url(tag_group=path_obj['tag_path'][x])
        url += _create_tags_url(tag=path_obj['tag_path'][len(path_obj['tag_path'])-1])
    except KeyError as err:
        err_msg = 'Error: No key {} identified | Function: {}'.format(err, inspect.currentframe().f_code.co_name)
        raise KepError(err_msg)
    except Exception as e:
        err_msg = 'Error: Error with {}: {}'.format(inspect.currentframe().f_code.co_name, str(e))
        raise KepError(err_msg)
    r = server._config_get(url)
    return r.payload

def get_all_tags(server, full_tag_path) -> list:
    '''Returns the properties of all "tag" object at a specific path in Kepware. 
    Returned object is JSON list.

    INPUTS:
    "server" - instance of the "server" class
    
    "full_tag_path" - path identifying location to retreive tag list. Standard Kepware address decimal 
    notation string including the tag such as "channel1.device1.tag_group1"
    
    RETURNS:
    list - data for the tags requested

    EXCEPTIONS:
    KepHTTPError - If urllib provides an HTTPError
    KepURLError - If urllib provides an URLError
    '''

    path_obj = kepconfig.path_split(full_tag_path)
    try:
        url = server.url+channel._create_url(path_obj['channel'])+device._create_url(path_obj['device'])
        if 'tag_path' in path_obj:
            for tg in path_obj['tag_path']:
                url += _create_tag_groups_url(tag_group=tg)
        url += _create_tags_url()
    except KeyError as err:
        err_msg = 'Error: No key {} identified | Function: {}'.format(err, inspect.currentframe().f_code.co_name)
        raise KepError(err_msg)
    except Exception as e:
        err_msg = 'Error: Error with {}: {}'.format(inspect.currentframe().f_code.co_name, str(e))
        raise KepError(err_msg)
    r = server._config_get(url)
    return r.payload

def get_tag_group(server, tag_group_path) -> dict:
    '''Returns the properties of the "tag group" object at a specific 
    path in Kepware. Returned object is JSON.

    INPUTS:
    "server" - instance of the "server" class 
    
    "tag_group_path" - path identifying tag group. Standard Kepware address decimal 
    notation string such as "channel1.device1.tag_group1"

    RETURNS:
    dict - data for the tag group requested

    EXCEPTIONS:
    KepHTTPError - If urllib provides an HTTPError
    KepURLError - If urllib provides an URLError
    '''
    path_obj = kepconfig.path_split(tag_group_path)
    try:
        url = server.url+channel._create_url(path_obj['channel'])+device._create_url(path_obj['device'])
        for tg in path_obj['tag_path']:
            url += _create_tag_groups_url(tag_group=tg)
    except KeyError as err:
        err_msg = 'Error: No key {} identified | Function: {}'.format(err, inspect.currentframe().f_code.co_name)
        raise KepError(err_msg)
    except Exception as e:
        err_msg = 'Error: Error with {}: {}'.format(inspect.currentframe().f_code.co_name, str(e))
        raise KepError(err_msg)
    r = server._config_get(url)
    return r.payload

def get_all_tag_groups(server, tag_group_path) -> list:
    '''Returns the properties of all "tag group" objects at a specific 
    path in Kepware. Returned object is JSON list. 

    INPUTS:
    "server" - instance of the "server" class
    
    "tag_group_path" - path identifying location to retreive tag group list. Standard Kepware address decimal 
    notation string such as "channel1.device1.tag_group1"

    RETURNS:
    list - data for the tag groups requested

    EXCEPTIONS:
    KepHTTPError - If urllib provides an HTTPError
    KepURLError - If urllib provides an URLError
    '''
    path_obj = kepconfig.path_split(tag_group_path)
    try:
        url = server.url+channel._create_url(path_obj['channel'])+device._create_url(path_obj['device'])
        if 'tag_path' in path_obj:
            for tg in path_obj['tag_path']:
                url += _create_tag_groups_url(tag_group=tg)
        url += _create_tag_groups_url()
    except KeyError as err:
        err_msg = 'Error: No key {} identified | Function: {}'.format(err, inspect.currentframe().f_code.co_name)
        raise KepError(err_msg)
    except Exception as e:
        err_msg = 'Error: Error with {}: {}'.format(inspect.currentframe().f_code.co_name, str(e))
        raise KepError(err_msg)
    r = server._config_get(url)
    return r.payload

def get_full_tag_structure(server, path, recursive = False) -> dict:
    '''Returns the properties of all "tag" and "tag group" objects at a specific 
    path in Kepware. Returned object is a dict of tag list and tag group list.

    Ex.
    {
        'tags': [tag1_dict, tag2_dict,...],
        'tag_groups':[tag_group1_dict, tag_group2_dict,...]
    } 

    If recursive is TRUE, then the call will iterate through all tag groups and get the tags and 
    tag groups of all tag group children.This would be the equivilant of asking for all tags and tag groups
    that exist below the "path" location. The returned object would look like below, nested based on how many 
    levels the tag_group namespace has tags or tag_groups:

    Ex.
    {
        'tags': [tag1_dict, tag2_dict,...],
        'tag_groups':[
            {
                tag_group1_properties,
                'tags': [tag1_dict, tag2_dict,...]
                'tag_groups':[sub_group1, subgroup2,...]
            }, 
            {
                tag_group2_properties,
                'tags': [tag1_dict, tag2_dict,...]
                'tag_groups':[sub_group1, subgroup2,...]
            },...]
    } 

    INPUTS:
    "server" - instance of the "server" class
    
    "path" - path identifying location to retreive the tag structure. Standard Kepware address decimal 
    notation string such as "channel1.device1.tag_group1" and must container at least the channel and device.

    RETURNS:
    dict - data for the tag structure requested at "path" location

    EXCEPTIONS:
    KepHTTPError - If urllib provides an HTTPError
    KepURLError - If urllib provides an URLError
    '''
    r = {}
        
    r['tags'] = get_all_tags(server, path)
    r['tag_groups'] = get_all_tag_groups(server, path)
    if recursive:
        for group in r['tag_groups']:
            res = get_full_tag_structure(server, path + '.' + group['common.ALLTYPES_NAME'])
            group.update(res)
    return r



# -------------------------------------------------------------------------
# Copyright (c) PTC Inc. and/or all its affiliates. All rights reserved.
# See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

# Package Setup Script - run this to build the package
# Usage - python setup.py <build> sdist bdist_wheel
#  Example: python setup.py 1.0.0 sdist bdist_wheel

import setuptools
import pathlib, os

def read(rel_path):
    # type: (str) -> str
    here = os.path.abspath(os.path.dirname(__file__))
    # intentionally *not* adding an encoding option to open, See:
    #   https://github.com/pypa/virtualenv/issues/201#issuecomment-3145690
    with open(os.path.join(here, rel_path)) as fp:
        return fp.read()

def get_version(rel_path):
    # type: (str) -> str
    for line in read(rel_path).splitlines():
        if line.startswith('__version__'):
            # __version__ = "0.9"
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    raise RuntimeError("Unable to find version string.")

here = pathlib.Path(__file__).parent.resolve()

# Get the long description from the README file
long_description = (here / 'README.md').read_text(encoding='utf-8')

setuptools.setup(
    name="kepconfig",
    version= get_version("kepconfig/__init__.py"),
    author="PTC Inc",
    description="SDK package for Kepware Configuration API",
    keywords="Kepware, OPC, Configuration, Thingworx",
    license="MIT License",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/PTCInc/Kepware-ConfigAPI-SDK-Python",
    project_urls={},
    packages=setuptools.find_packages(),
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
        "Intended Audience :: Manufacturing",
        "Intended Audience :: Developers",
    ],
    python_requires='>=3.6',
)



# -------------------------------------------------------------------------
# Copyright (c) 2020, PTC Inc. and/or all its affiliates. All rights reserved.
# See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

__version__ = "1.0.0"
__path__ = __import__("pkgutil").extend_path(__path__, __name__)
from . import connection, error

def path_split(path):
    '''Used to split the standard Kepware address decimal notation into a dict that contains the 
    "channel", "device" and "tag_path".

    Ex. path input = "ch1.dev1.tg1.tg2.tg3"

    Returns: {'channel': 'ch1', 'device': 'dev1', 'tag_path': ['tg1','tg2','tg3']}

    Ex. path input = "ch1.dev1"

    Returns: {'channel': 'ch1', 'device': 'dev1'}
    '''
    path_list = path.split('.', 2)
    path_obj = {}
    for x in range(0, len(path_list)):
        if x == 0:
            path_obj['channel'] = path_list[0]
        elif x == 1:
            path_obj['device'] = path_list[1]
        else:
            path_obj['tag_path'] = path_list[2].split('.')
    return path_obj

def _address_dedecimal(tag_address):
    if tag_address[0] is '_':
        tag_address = tag_address[1::]
    updated = tag_address.replace('.','_')
    return updated